# Generated by Django 3.2.19 on 2024-05-24 12:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0060_workspace_chain'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='supersedes',
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='superseded',
                to='db.workrequest',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_workers."""

from debusine.db.models import Token, Worker
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListWorkersCommandTests(TabularOutputTests, TestCase):
    """Test for list_workers management command."""

    def test_list_workers_connected(self) -> None:
        """
        List worker command prints worker information.

        The worker is connected.
        """
        token = Token.objects.create()
        worker_1 = Worker.objects.create_with_fqdn('recent-ping', token)
        worker_1.mark_connected()
        assert worker_1.connected_at is not None

        with self.assertPrintsTable() as output:
            stdout, stderr, _ = call_command('list_workers')
        self.assertEqual(output.col(0), [worker_1.name])
        self.assertEqual(output.col(2), [worker_1.registered_at.isoformat()])
        self.assertEqual(output.col(3), [worker_1.connected_at.isoformat()])
        self.assertEqual(output.col(4), [token.hash])
        self.assertEqual(output.col(5), [str(token.enabled)])

    def test_list_workers_not_connected(self) -> None:
        """
        List worker command prints worker information.

        The worker is not connected.
        """
        token = Token.objects.create()
        worker_1 = Worker.objects.create_with_fqdn('recent-ping', token=token)
        with self.assertPrintsTable() as output:
            stdout, stderr, _ = call_command('list_workers')

        self.assertEqual(output.col(0), [worker_1.name])
        self.assertEqual(output.col(2), [worker_1.registered_at.isoformat()])
        self.assertEqual(output.col(3), ["-"])
        self.assertEqual(output.col(4), [token.hash])
        self.assertEqual(output.col(5), [str(token.enabled)])

    def test_list_workers_celery(self) -> None:
        """list_workers handles Celery workers, which have no tokens."""
        worker = Worker.objects.get_or_create_celery()
        worker.mark_connected()
        assert worker.connected_at is not None
        with self.assertPrintsTable() as output:
            call_command("list_workers")

        self.assertEqual(output.col(0), [worker.name])
        self.assertEqual(output.col(1), [worker.worker_type])
        self.assertEqual(output.col(2), [worker.registered_at.isoformat()])
        self.assertEqual(output.col(3), [worker.connected_at.isoformat()])
        self.assertEqual(output.col(4), ["-"])
        self.assertEqual(output.col(5), ["-"])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Migrate WorkRequest debusine:signing-key artifact lookups to fingerprints."""

import logging
import re
from dataclasses import dataclass

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from debusine.tasks.models import TaskTypes

log = logging.getLogger(__name__)


@dataclass
class TaskMigration:
    task_type: TaskTypes
    task_name: str
    data_key: str
    dynamic_data_key: str | None
    # For workflows, we have to find the dynamic data in a child
    child_task_type: TaskTypes | None
    child_task_name: str | None
    child_data_key: str | None


# Ordered bottom-up, topologically, so that we can determine workflow keys from
# immediate children
task_migrations = [
    # Leaf tasks
    TaskMigration(
        task_type=TaskTypes.SIGNING,
        task_name="sign",
        data_key="key",
        dynamic_data_key="key_id",
        child_task_type=None,
        child_task_name=None,
        child_data_key=None,
    ),
    TaskMigration(
        task_type=TaskTypes.SIGNING,
        task_name="debsign",
        data_key="key",
        dynamic_data_key="key_id",
        child_task_type=None,
        child_task_name=None,
        child_data_key=None,
    ),
    # Level 0 workflows
    TaskMigration(
        task_type=TaskTypes.WORKFLOW,
        task_name="make_signed_source",
        data_key="key",
        dynamic_data_key=None,
        child_task_type=TaskTypes.SIGNING,
        child_task_name="sign",
        child_data_key="key",
    ),
    TaskMigration(
        task_type=TaskTypes.WORKFLOW,
        task_name="package_upload",
        data_key="key",
        dynamic_data_key=None,
        child_task_type=TaskTypes.SIGNING,
        child_task_name="debsign",
        child_data_key="key",
    ),
    # Level 1 workflows
    TaskMigration(
        task_type=TaskTypes.WORKFLOW,
        task_name="debian_pipeline",
        data_key="make_signed_source_key",
        dynamic_data_key=None,
        child_task_type=TaskTypes.WORKFLOW,
        child_task_name="make_signed_source",
        child_data_key="key",
    ),
    TaskMigration(
        task_type=TaskTypes.WORKFLOW,
        task_name="debian_pipeline",
        data_key="upload_key",
        dynamic_data_key=None,
        child_task_type=TaskTypes.WORKFLOW,
        child_task_name="package_upload",
        child_data_key="key",
    ),
]

artifact_id_lookup_re = re.compile(r"^(\d+)@artifacts$")


def migrate_to_fingerprints(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Migrate references to debusine:signing-key artifacts to fingerprints."""
    Artifact = apps.get_model("db", "Artifact")
    WorkRequest = apps.get_model("db", "WorkRequest")

    for step in task_migrations:
        for work_request in WorkRequest.objects.filter(
            task_type=step.task_type,
            task_name=step.task_name,
            task_data__has_key=step.data_key,
        ).order_by("id"):
            lookup: int | str | None = work_request.task_data[step.data_key]
            if lookup is None:
                continue
            artifact_id: int | None = None
            if isinstance(lookup, int):
                artifact_id = lookup
            elif lookup.isdigit():
                artifact_id = int(lookup)
            elif m := artifact_id_lookup_re.match(lookup):
                artifact_id = int(m.group(1))
            elif (
                step.dynamic_data_key
                and step.dynamic_data_key in work_request.dynamic_task_data
            ):
                artifact_id = work_request.dynamic_task_data[
                    step.dynamic_data_key
                ]

            if artifact_id is not None:
                try:
                    artifact = Artifact.objects.get(
                        id=artifact_id, category="debusine:signing-key"
                    )
                except Artifact.DoesNotExist:
                    log.warning(
                        (
                            "Unable to migrate '%s' to fingerprints in "
                            "work request %i, debusine:signing-key "
                            "artifact %i not found."
                        ),
                        step.data_key,
                        work_request.id,
                        artifact_id,
                    )
                    continue
                work_request.task_data[step.data_key] = artifact.data[
                    "fingerprint"
                ]
                if step.dynamic_data_key and work_request.dynamic_task_data:
                    work_request.dynamic_task_data.pop(step.dynamic_data_key)
                work_request.save()
                continue

            elif step.child_task_type:
                child = (
                    work_request.children.filter(
                        task_type=step.child_task_type,
                        task_name=step.child_task_name,
                        task_data__has_key=step.child_data_key,
                    )
                    .order_by("id")
                    .first()
                )
                if child is None:
                    log.warning(
                        (
                            "Unable to migrate '%s' to fingerprints in "
                            "work request %i, child work request with "
                            "task_name '%s' and fingerprints not found."
                        ),
                        step.data_key,
                        work_request.id,
                        step.child_task_name,
                    )
                    continue
                work_request.task_data[step.data_key] = child.task_data[
                    step.child_data_key
                ]
                # No workflow tasks have dynamic data, but included for
                # logical completeness:
                if step.dynamic_data_key:  # pragma: no cover
                    work_request.dynamic_task_data.pop(step.dynamic_data_key)
                work_request.save()
                continue

            log.warning(
                "Unable to migrate '%s' to fingerprints in work request %i.",
                step.data_key,
                work_request.id,
            )


def migrate_to_signing_key_artifacts(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Migrate fingerprints back to debusine:signing-key asset lookups."""
    Artifact = apps.get_model("db", "Artifact")
    WorkRequest = apps.get_model("db", "WorkRequest")

    for step in task_migrations:
        for work_request in WorkRequest.objects.filter(
            task_type=step.task_type,
            task_name=step.task_name,
            task_data__has_key=step.data_key,
        ).order_by("id"):
            fingerprint = work_request.task_data[step.data_key]
            artifact = (
                Artifact.objects.filter(
                    category="debusine:signing-key",
                    data__fingerprint=fingerprint,
                )
                .order_by("id")
                .first()
            )
            if artifact is None:
                log.warning(
                    (
                        "Unable to migrate '%s' to artifact lookups in work "
                        "request %i, artifact not found"
                    ),
                    step.data_key,
                    work_request.id,
                )
                continue
            work_request.task_data[step.data_key] = f"{artifact.id}@artifacts"
            if step.dynamic_data_key:
                work_request.dynamic_task_data[step.dynamic_data_key] = (
                    artifact.id
                )
            work_request.save()


class Migration(migrations.Migration):

    dependencies = [
        ("db", "0116_assets"),
    ]

    operations = [
        migrations.RunPython(
            migrate_to_fingerprints, migrate_to_signing_key_artifacts
        )
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for cloud-init script generation."""

from base64 import b64decode
from textwrap import dedent

import yaml
from django.test import SimpleTestCase

from debusine.db.models import Token
from debusine.server.worker_pools.cloud_init import (
    AptConfig,
    CloudInitConfig,
    FilesystemSetup,
    WriteFile,
    worker_bootstrap_cloud_init,
)
from debusine.server.worker_pools.models import (
    DebianRelease,
    DebusineInstallSource,
)
from debusine.test.django import TestCase


class CloudInitConfigTest(SimpleTestCase):
    """Tests for CloudInitConfig."""

    def test_full_instantiation(self) -> None:
        config = CloudInitConfig(
            apt=AptConfig(
                preserve_sources_list=False,
                sources_list="deb $MIRROR $RELEASE main",
            ),
            device_aliases={"foo": "/dev/nvme0n1"},
            fs_setup=[
                FilesystemSetup(device="foo", filesystem="ext4"),
            ],
            mounts=[
                ["foo", "/mnt/foo", "ext4", "defaults"],
            ],
            package_reboot_if_required=True,
            package_update=True,
            package_upgrade=True,
            packages=["build-essential", "vim-tiny"],
            bootcmd=["/bin/true"],
            runcmd=["/bin/true", "/bin/false"],
            write_files=[
                WriteFile(
                    path="/etc/hostname", content="test", permissions="0o644"
                ),
            ],
        )
        rendered = config.as_str()
        parsed = yaml.safe_load(rendered)
        self.assertIsInstance(parsed, dict)
        self.assertEqual(
            parsed["apt"],
            {
                "preserve_sources_list": False,
                "sources_list": "deb $MIRROR $RELEASE main",
            },
        )

        self.assertEqual(
            parsed["fs_setup"], [{"device": "foo", "filesystem": "ext4"}]
        )
        self.assertTrue(parsed["package_reboot_if_required"])

    def test_empty_as_str(self) -> None:
        config = CloudInitConfig()
        self.assertEqual(config.as_str(), "#cloud-config\n{}\n")

    def test_simple_as_str(self) -> None:
        config = CloudInitConfig(packages=["foobar"], runcmd=["/bin/true"])
        self.assertEqual(
            config.as_str(),
            dedent(
                """\
                #cloud-config
                packages:
                - foobar
                runcmd:
                - /bin/true
                """
            ),
        )

    def test_simple_as_base64_str(self) -> None:
        config = CloudInitConfig(packages=["foobar"], runcmd=["/bin/true"])
        encoded = config.as_base64_str()
        parsed = yaml.safe_load(b64decode(encoded))
        self.assertEqual(
            parsed, {"packages": ["foobar"], "runcmd": ["/bin/true"]}
        )


class TestWorkerBootstrapCloudInit(TestCase):
    """Tests for worker_bootstrap_cloud_init."""

    def test_worker_bootstrap_cloud_init_simple(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.BACKPORTS, None
        )
        self.assertTrue(config.package_update)
        self.assertTrue(config.package_upgrade)
        self.assertTrue(config.package_reboot_if_required)
        self.assertEqual(len(config.bootcmd), 1)
        self.assertIn("eatmydata", config.bootcmd[0])
        self.assertIn("debusine-worker", config.packages)
        self.assertIn("incus", config.packages)
        apt_pin_files = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/preferences.d/90-debusine"
        ]
        self.assertEqual(len(apt_pin_files), 1)
        self.assertIn("src:debusine", apt_pin_files[0].content)
        self.assertIn("a=stable-backports", apt_pin_files[0].content)
        token_files = [
            file
            for file in config.write_files
            if file.path == "/etc/debusine/worker/activation-token"
        ]
        self.assertEqual(len(token_files), 1)
        self.assertEqual(token_files[0].content, token.key)
        self.assertIsNone(token_files[0].owner)
        nftables_files = [
            file
            for file in config.write_files
            if file.path == "/etc/nftables.conf"
        ]
        self.assertEqual(len(nftables_files), 1)
        daily_build_sources = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/sources.list.d/debusine.sources"
        ]
        self.assertEqual(len(daily_build_sources), 0)

    def test_worker_bootstrap_cloud_init_daily(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.DAILY_BUILDS, None
        )
        apt_pin_files = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/preferences.d/90-debusine"
        ]
        self.assertEqual(len(apt_pin_files), 0)
        daily_build_sources = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/sources.list.d/debusine.sources"
        ]
        self.assertEqual(len(daily_build_sources), 1)

    def test_worker_bootstrap_cloud_init_pre_installed(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.PRE_INSTALLED, None
        )
        self.assertTrue(config.package_update)
        self.assertTrue(config.package_upgrade)
        self.assertTrue(config.package_reboot_if_required)
        self.assertEqual(config.bootcmd, [])
        self.assertEqual(
            config.write_files,
            [
                WriteFile(
                    path="/etc/debusine/worker/activation-token",
                    permissions="0o600",
                    content=token.key,
                    owner="debusine-worker:debusine-worker",
                )
            ],
        )
        self.assertEqual(
            config.runcmd, ["systemctl restart debusine-worker.service"]
        )
        self.assertEqual(config.packages, [])

    def test_worker_bootstrap_bookworm_bpo(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.BACKPORTS, DebianRelease.BOOKWORM
        )
        apt_pin_files = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/preferences.d/90-debusine"
        ]
        self.assertEqual(len(apt_pin_files), 1)
        self.assertIn("src:debusine", apt_pin_files[0].content)
        self.assertIn("src:python-django", apt_pin_files[0].content)
        self.assertIn("n=bookworm-backports", apt_pin_files[0].content)

    def test_worker_bootstrap_bookworm_daily(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.DAILY_BUILDS, DebianRelease.BOOKWORM
        )
        self.assertIsNotNone(config.apt)
        assert config.apt
        self.assertFalse(config.apt.preserve_sources_list)
        self.assertEqual(
            config.apt.sources_list,
            dedent(
                """\
                deb $MIRROR $RELEASE main contrib non-free-firmware
                deb $MIRROR $RELEASE-backports main contrib non-free-firmware
                deb $SECURITY $RELEASE-security main contrib non-free-firmware
                """
            ),
        )
        apt_pin_files = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/preferences.d/90-debusine"
        ]
        self.assertEqual(len(apt_pin_files), 1)
        self.assertNotIn("src:debusine", apt_pin_files[0].content)
        self.assertIn("src:python-django", apt_pin_files[0].content)
        self.assertIn("n=bookworm-backports", apt_pin_files[0].content)
        daily_build_sources = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/sources.list.d/debusine.sources"
        ]
        self.assertEqual(len(daily_build_sources), 1)

    def test_worker_bootstrap_trixie(self) -> None:
        token = self.playground.create_bare_token()
        config = worker_bootstrap_cloud_init(
            token, DebusineInstallSource.RELEASE, DebianRelease.TRIXIE
        )
        self.assertIsNone(config.apt)
        apt_pin_files = [
            file
            for file in config.write_files
            if file.path == "/etc/apt/preferences.d/90-debusine"
        ]
        self.assertEqual(len(apt_pin_files), 0)

    def test_worker_bootstrap_cloud_init_without_key(self) -> None:
        token = Token()
        with self.assertRaisesRegex(
            ValueError,
            (
                r"Requires a fresh token with a key. Token<None> does not have "
                r"one\."
            ),
        ):
            worker_bootstrap_cloud_init(
                token, DebusineInstallSource.BACKPORTS, None
            )



# Generated by Django 3.2.19 on 2024-05-22 10:05

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0059_collection_workspace_namespace'),
    ]

    operations = [
        migrations.CreateModel(
            name='WorkspaceChain',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'order',
                    models.IntegerField(
                        help_text='Lookup order of this element in the chain'
                    ),
                ),
                (
                    'child',
                    models.ForeignKey(
                        help_text='Workspace that falls back on `parent` for lookups',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='chain_parents',
                        to='db.workspace',
                    ),
                ),
                (
                    'parent',
                    models.ForeignKey(
                        help_text='Workspace to be looked up if lookup in `child` fails',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='chain_children',
                        to='db.workspace',
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name='workspace',
            name='inherits',
            field=models.ManyToManyField(
                related_name='inherited_by',
                through='db.WorkspaceChain',
                to='db.Workspace',
            ),
        ),
        migrations.AddConstraint(
            model_name='workspacechain',
            constraint=models.UniqueConstraint(
                fields=('child', 'parent'),
                name='db_workspacechain_unique_child_parent',
            ),
        ),
        migrations.AddConstraint(
            model_name='workspacechain',
            constraint=models.UniqueConstraint(
                fields=('child', 'order'),
                name='db_workspacechain_unique_child_order',
            ),
        ),
    ]



# Generated by Django 4.2.15 on 2024-12-13 16:34

from django.db import migrations

import debusine.db.models.auth


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0113_workrequest_output_data_json'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='user',
            managers=[
                ('objects', debusine.db.models.auth.UserManager()),
            ],
        ),
    ]



# Generated by Django 3.2.19 on 2024-04-29 16:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0041_piuparts_lookup'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='is_system',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='user',
            name='email',
            field=models.EmailField(
                blank=True, max_length=254, verbose_name='email address'
            ),
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('is_system', False), ('email', ''), _connector='OR'
                ),
                name='db_user_non_system_email',
            ),
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.UniqueConstraint(
                condition=models.Q(('is_system', False)),
                fields=('email',),
                name='db_user_unique_email',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the artifact views."""

import json
import os.path
import unittest
from typing import ClassVar, NamedTuple, cast
from unittest import mock

from django.http.request import HttpHeaders
from django.http.response import HttpResponseBase, HttpResponseRedirectBase
from django.template import engines
from django.test import override_settings
from django.utils.datastructures import CaseInsensitiveMapping
from rest_framework import status

from debusine.db.context import context
from debusine.db.models import File, FileInArtifact
from debusine.server.views import ProblemResponse
from debusine.test.django import TestCase
from debusine.web.views import files
from debusine.web.views.files import (
    FileDownloadMixin,
    FileUI,
    FileWidget,
    MAX_FILE_SIZE,
    PathMixin,
)
from debusine.web.views.tests.utils import ViewTestMixin


class FileUITests(TestCase):
    """Tests for the FileUI class."""

    def assertRenders(
        self, file_ui: FileUI, file_in_artifact: FileInArtifact
    ) -> None:
        """Test that the given widget renders without errors."""
        template = engines["django"].from_string(
            "{% load debusine %}{% widget widget %}"
        )
        setattr(template, "engine", engines["django"])
        widget = file_ui.widget_class(file_in_artifact, file_ui=file_ui)
        # Set DEBUG=True to have render raise exceptions instead of logging
        # them
        with override_settings(DEBUG=True):
            template.render({"widget": widget})

    def test_from_file_in_artifact(self) -> None:
        """Test from_file_in_artifact."""

        class SubTest(NamedTuple):
            """Define a subtest."""

            filename: str
            widget_class: type[FileWidget]
            content_type: str
            size: int = 1024

        sub_tests = [
            SubTest(
                "build.changes",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.log", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "file.txt", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "hello.build", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "hello.buildinfo",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.sources",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "readme.md",
                files.TextFileWidget,
                "text/markdown; charset=utf-8",
            ),
            SubTest(
                "a.out", files.BinaryFileWidget, "application/octet-stream"
            ),
            SubTest(
                "big.bin",
                files.TooBigFileWidget,
                "application/octet-stream",
                2**32,
            ),
        ]

        for sub_test in sub_tests:
            with self.subTest(sub_test.filename):
                with context.disable_permission_checks():
                    artifact, _ = self.playground.create_artifact(
                        paths={sub_test.filename: b"test"}, create_files=True
                    )
                file_in_artifact = FileInArtifact.objects.get(artifact=artifact)
                file_in_artifact.file.size = sub_test.size
                fileui = FileUI.from_file_in_artifact(file_in_artifact)
                self.assertEqual(fileui.content_type, sub_test.content_type)
                self.assertEqual(fileui.widget_class, sub_test.widget_class)
                self.assertRenders(fileui, file_in_artifact)

    def test_compressed(self) -> None:
        """Test from_file_in_artifact with a compressed file."""
        file_in_artifact = FileInArtifact(path="file.md.gz")
        file_in_artifact.file = File(size=123)
        fileui = FileUI.from_file_in_artifact(file_in_artifact)
        self.assertEqual(fileui.content_type, "application/gzip")
        self.assertEqual(fileui.widget_class, files.BinaryFileWidget)


class PathMixinTests(ViewTestMixin, unittest.TestCase):
    """Tests for the ArtifactDetailView class."""

    def test_normalize_path(self) -> None:
        """Test PathMixin.normalize_path."""
        f = PathMixin.normalize_path
        self.assertEqual(f(""), "/")
        self.assertEqual(f("/"), "/")
        self.assertEqual(f("."), "/")
        self.assertEqual(f(".."), "/")
        self.assertEqual(f("../"), "/")
        self.assertEqual(f("../../.././../../"), "/")
        self.assertEqual(f("src/"), "/src/")
        self.assertEqual(f("src/.."), "/")
        self.assertEqual(f("/a/b/../c/./d//e/f/../g/"), "/a/c/d/e/g/")

    def test_path(self) -> None:
        """Test PathMixin.path."""
        view = self.instantiate_view_class(PathMixin, "/")
        self.assertEqual(view.path, "")

        view = self.instantiate_view_class(
            PathMixin, "/", path="/foo/bar/../baz"
        )
        self.assertEqual(view.path, "foo/baz/")


class FileDownloadMixinTests(ViewTestMixin, TestCase):
    """Test FileDownloadMixin."""

    playground_memory_file_store = False

    contents: ClassVar[dict[str, bytes]]
    empty: ClassVar[FileInArtifact]
    file: ClassVar[FileInArtifact]

    @classmethod
    @context.disable_permission_checks()
    def setUpTestData(cls) -> None:
        """Set up the common test fixture."""
        super().setUpTestData()
        cls.contents = {
            "empty.bin": b"",
            "file.md": bytes(range(256)),
        }
        artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.empty = artifact.fileinartifact_set.get(path="empty.bin")
        cls.file = artifact.fileinartifact_set.get(path="file.md")

    def get_stream_response(
        self,
        file_in_artifact: FileInArtifact,
        range_header: tuple[int, int] | str | None = None,
        download: bool = True,
    ) -> HttpResponseBase:
        """Instantiate the view and get a streaming file response."""
        request = self.make_request("/")
        headers = {}
        match range_header:
            case str():
                headers["Range"] = range_header
            case [range_start, range_end]:
                headers["Range"] = f"bytes={range_start}-{range_end}"
        if headers:
            request.headers = cast(
                HttpHeaders,
                CaseInsensitiveMapping({**request.headers, **headers}),
            )

        ui_info = FileUI.from_file_in_artifact(file_in_artifact)
        view = self.instantiate_view_class(FileDownloadMixin, request)
        response = view.stream_file(file_in_artifact, ui_info, download)
        if isinstance(response, ProblemResponse):
            # This is needed by assertResponseProblem
            setattr(
                response,
                "json",
                lambda: json.loads(response.content.decode(response.charset)),
            )

        return response

    def assertFileResponse(
        self,
        response: HttpResponseBase,
        file_in_artifact: FileInArtifact,
        range_start: int = 0,
        range_end: int | None = None,
        disposition: str = "attachment",
    ) -> None:
        """Assert that response has the expected headers and content."""
        if range_start == 0 and range_end is None:
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        else:
            self.assertEqual(
                response.status_code, status.HTTP_206_PARTIAL_CONTENT
            )

        file_contents = self.contents[file_in_artifact.path]
        if range_end is None:
            range_end = len(file_contents) - 1
        expected_contents = file_contents[range_start : range_end + 1]

        headers = response.headers
        self.assertEqual(headers["Accept-Ranges"], "bytes")
        self.assertEqual(headers["Content-Length"], str(len(expected_contents)))

        if len(expected_contents) > 0:
            self.assertEqual(
                headers["Content-Range"],
                f"bytes {range_start}-{range_end}/{len(file_contents)}",
            )

        filename = os.path.basename(file_in_artifact.path)
        self.assertEqual(
            headers["Content-Disposition"],
            f'{disposition}; filename="{filename}"',
        )

        streaming_content = getattr(response, "streaming_content", None)
        assert streaming_content is not None
        self.assertEqual(b"".join(streaming_content), expected_contents)

    def test_get_file(self) -> None:
        """Get return the file."""
        response = self.get_stream_response(self.file)
        self.assertFileResponse(response, self.file)
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_file_inline(self) -> None:
        """Get return the file."""
        response = self.get_stream_response(self.file, download=False)
        self.assertFileResponse(response, self.file, disposition="inline")
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_empty_file(self) -> None:
        """Test empty downloadable file (which mmap doesn't support)."""
        response = self.get_stream_response(self.empty)
        self.assertFileResponse(response, self.empty)
        self.assertEqual(
            response.headers["content-type"],
            "application/octet-stream",
        )

    def test_get_incomplete_file(self) -> None:
        """Get returns 404 for an incomplete file."""
        self.file.complete = False
        self.file.save()
        response = self.get_stream_response(self.file)
        self.assertResponseProblem(
            response,
            "Cannot download incomplete file",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_get_file_range(self) -> None:
        """Get return part of the file (based on Range header)."""
        start, end = 10, 20
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_content_range_to_end_of_file(self) -> None:
        """Server returns a file from a position to the end."""
        start, end = 5, len(self.contents["file.md"]) - 1
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_content_range_invalid(self) -> None:
        """Get return an error: Range header was invalid."""
        invalid_range_header = "invalid-range"
        response = self.get_stream_response(
            self.file, range_header=invalid_range_header
        )
        self.assertResponseProblem(
            response, f'Invalid Range header: "{invalid_range_header}"'
        )

    def test_get_file_range_start_greater_file_size(self) -> None:
        """Get return 400: client requested an invalid start position."""
        file_size = len(self.contents["file.md"])
        start, end = file_size + 10, file_size + 20
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range start: {start}. " f"File size: {file_size}",
        )

    def test_get_file_range_end_is_file_size(self) -> None:
        """Get return 400: client requested and invalid end position."""
        end = len(self.contents["file.md"])
        response = self.get_stream_response(self.file, range_header=(0, end))
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range end: {end}. File size: {end}",
        )

    def test_get_file_range_end_greater_file_size(self) -> None:
        """Get return 400: client requested an invalid end position."""
        file_size = len(self.contents["file.md"])
        end = file_size + 10
        response = self.get_stream_response(self.file, range_header=(0, end))
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range end: {end}. " f"File size: {file_size}",
        )

    def test_get_file_url_redirect(self) -> None:
        """
        Get file response: redirect if get_url for the file is available.

        This would happen if the file is stored in a FileStore supporting
        get_url (e.g. an object storage) instead of being served from the
        server's file system.
        """
        destination_url = "https://some-backend.net/file?token=asdf"

        with mock.patch(
            "debusine.server.file_backend.local.LocalFileBackendEntry.get_url",
            autospec=True,
            return_value=destination_url,
        ) as mock_get_url:
            response = self.get_stream_response(self.file)
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            assert isinstance(response, HttpResponseRedirectBase)
            self.assertEqual(response.url, destination_url)
            ui_info = FileUI.from_file_in_artifact(self.file)
            mock_get_url.assert_called_once_with(
                mock.ANY, content_type=ui_info.content_type
            )


class FileWidgetTests(ViewTestMixin, TestCase):
    """Test FileView."""

    contents: ClassVar[dict[str, bytes]]
    binary: ClassVar[FileInArtifact]
    dsc: ClassVar[FileInArtifact]
    empty: ClassVar[FileInArtifact]
    large: ClassVar[FileInArtifact]
    text: ClassVar[FileInArtifact]

    @classmethod
    @context.disable_permission_checks()
    def setUpTestData(cls) -> None:
        """Set up the common test fixture."""
        super().setUpTestData()
        cls.contents = {
            "file.md": b"# Title\nText",
            "file.dsc": b"Source: hello",
            "empty.bin": b"",
            "file.bin": bytes(range(256)),
            "largefile.bin": b"large",
        }
        artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.empty = artifact.fileinartifact_set.get(path="empty.bin")
        cls.dsc = artifact.fileinartifact_set.get(path="file.dsc")
        cls.text = artifact.fileinartifact_set.get(path="file.md")
        cls.binary = artifact.fileinartifact_set.get(path="file.bin")
        cls.large = artifact.fileinartifact_set.get(path="largefile.bin")
        cls.large.file.size = 2**32
        cls.large.file.save()

    def test_context_data_text(self) -> None:
        """Test context for text files."""
        file = self.text
        context = FileWidget.create(file).get_context_data()

        file_linenumbers = context.pop("file_linenumbers")
        self.assertIn("L1", file_linenumbers)
        file_content = context.pop("file_content")
        self.assertIn("# Title", file_content)

        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "file_contents_div_id": "file-contents",
            },
        )

    def test_context_data_text_incomplete(self) -> None:
        """Test context for incomplete text files."""
        file = self.text
        file.complete = False
        file.save()
        context = FileWidget.create(file).get_context_data()

        self.assertNotIn("file_linenumbers", context)
        self.assertNotIn("file_content", context)

        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "file_contents_div_id": "file-contents",
            },
        )

    def test_context_data_dsc(self) -> None:
        """Test context for text files."""
        file = self.dsc
        context = FileWidget.create(file).get_context_data()

        file_linenumbers = context.pop("file_linenumbers")
        self.assertIn("L1", file_linenumbers)
        file_content = context.pop("file_content")
        self.assertIn('<span class="k">Source</span>', file_content)

        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "file_contents_div_id": "file-contents",
            },
        )

    def test_context_data_empty(self) -> None:
        """Test context for empty binary files."""
        file = self.empty
        context = FileWidget.create(file).get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
            },
        )

    def test_context_data_binary(self) -> None:
        """Test context for binary files."""
        file = self.binary
        context = FileWidget.create(file).get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
            },
        )

    def test_context_data_large(self) -> None:
        """Test context for files too large."""
        file = self.large
        context = FileWidget.create(file).get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "file_max_size": MAX_FILE_SIZE,
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for base signing task classes."""

from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from django.db import connections
from django.test import TestCase, TransactionTestCase

from debusine.signing.tasks import BaseSigningTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData, WorkerType
from debusine.tasks.tests.helper_mixin import TestBaseTask


class TestBaseSigningTask(
    TestBaseTask[BaseTaskData, BaseDynamicTaskData],
    BaseSigningTask[BaseTaskData, BaseDynamicTaskData],
):
    """Sample class to test BaseSigningTask class."""

    def run(self, execute_directory: Path) -> bool:  # noqa: U100
        """Succeed if this task is running in a transaction."""
        return connections["default"].in_atomic_block


class TestBaseSigningTask2Data(BaseTaskData):
    """Data representation for TestBaseSigningTask2."""

    __test__ = False

    foo: str


class TestBaseSigningTask2(
    TestBaseTask[TestBaseSigningTask2Data, BaseDynamicTaskData],
    BaseSigningTask[TestBaseSigningTask2Data, BaseDynamicTaskData],
):
    """Test BaseSigningTask class with jsonschema validation."""

    TASK_VERSION = 1

    def run(self, execute_directory: Path) -> bool:  # noqa: U100
        """Unused abstract method from BaseExternalTask."""
        raise NotImplementedError()


class BaseSigningTaskTests(TestCase):
    """Unit tests for :class:`BaseSigningTask`."""

    def setUp(self) -> None:
        """Create the shared attributes."""
        self.task = TestBaseSigningTask({})
        self.task2 = TestBaseSigningTask2({"foo": "bar"})
        self.worker_metadata = {"system:worker_type": WorkerType.SIGNING}

    def test_can_run_on_no_version(self) -> None:
        """Ensure can_run_on returns True if no version is specified."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        self.assertEqual(self.task.can_run_on(metadata), True)

    def test_can_run_on_with_different_versions(self) -> None:
        """Ensure can_run_on returns False if versions differ."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        metadata["signing:testbasesigningtask:version"] = 1
        self.assertEqual(self.task.can_run_on(metadata), False)


class BaseSigningTaskTransactionTests(TransactionTestCase):
    """Test transactional behaviour of :class:`BaseSigningTask`."""

    def test_runs_in_transaction(self) -> None:
        """Tasks are executed in a transaction."""
        task = TestBaseSigningTask({}, {})
        task._debug_log_files_directory = TemporaryDirectory(
            prefix="debusine-tests-"
        )
        self.addCleanup(task._debug_log_files_directory.cleanup)
        with mock.patch.object(
            task, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertTrue(task.execute())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Database interaction for worker tasks on the server.

This TaskDatabase extension is used when Task code is run server-side with
database access.
"""

from collections.abc import Collection as AbcCollection
from typing import overload

from django.conf import settings

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import LookupChildType, RelationType
from debusine.db.models.artifacts import Artifact
from debusine.db.models.work_requests import WorkRequest
from debusine.tasks.models import LookupMultiple, LookupSingle
from debusine.tasks.server import (
    ArtifactInfo,
    MultipleArtifactInfo,
    TaskDatabaseInterface,
)


class TaskDatabase(TaskDatabaseInterface):
    """Implementation of database interaction in worker tasks."""

    def __init__(self, work_request: WorkRequest) -> None:
        """Construct a :py:class:`TaskDatabase`."""
        self.work_request = work_request

    @staticmethod
    def _make_artifact_info(artifact: "Artifact") -> ArtifactInfo:
        """Extract basic information from an artifact."""
        return ArtifactInfo(
            id=artifact.id,
            category=artifact.category,
            data=artifact.create_data(),
        )

    @overload
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo: ...

    @overload
    def lookup_single_artifact(
        self,
        lookup: None,
        default_category: CollectionCategory | None = None,
    ) -> None: ...

    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo | None:
        """
        Look up a single artifact.

        :param lookup: A :ref:`lookup-single`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about the artifact, or None if the provided
          lookup is None (for convenience in some call sites).
        :raises KeyError: if the lookup does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way.
        """
        # Import here to prevent circular imports
        from debusine.server.collections.lookup import lookup_single

        return (
            None
            if lookup is None
            else self._make_artifact_info(
                lookup_single(
                    lookup=lookup,
                    workspace=self.work_request.workspace,
                    user=self.work_request.created_by,
                    default_category=default_category,
                    workflow_root=self.work_request.get_workflow_root(),
                    expect_type=LookupChildType.ARTIFACT,
                ).artifact
            )
        )

    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple | None,
        default_category: CollectionCategory | None = None,
    ) -> MultipleArtifactInfo:
        """
        Look up multiple artifacts.

        :param lookup: A :ref:`lookup-multiple`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about each artifact.
        :raises KeyError: if any of the lookups does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way.
        """
        # Import here to prevent circular imports
        from debusine.server.collections.lookup import lookup_multiple

        return MultipleArtifactInfo(
            []
            if lookup is None
            else [
                self._make_artifact_info(result.artifact)
                for result in lookup_multiple(
                    lookup=lookup,
                    workspace=self.work_request.workspace,
                    user=self.work_request.created_by,
                    default_category=default_category,
                    workflow_root=self.work_request.get_workflow_root(),
                    expect_type=LookupChildType.ARTIFACT,
                )
            ]
        )

    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
    ) -> MultipleArtifactInfo:
        """
        Find artifacts via relations.

        For each of the artifacts in ``artifact_ids``, find all artifacts of
        ``target_category`` that are related to the initial artifacts with a
        relation of type ``relation_type``.

        :param artifact_ids: The IDs of artifacts to start from.
        :param target_category: The category of artifacts to find.
        :param relation_type: The type of relation to follow.
        :return: Information about each artifact.
        """
        return MultipleArtifactInfo(
            [
                self._make_artifact_info(artifact)
                for artifact in Artifact.objects.filter(
                    targeted_by__artifact__in=artifact_ids,
                    category=target_category,
                    targeted_by__type=relation_type,
                )
            ]
        )

    @overload
    def lookup_single_collection(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> int: ...

    @overload
    def lookup_single_collection(
        self,
        lookup: None,
        default_category: CollectionCategory | None = None,
    ) -> None: ...

    def lookup_single_collection(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> int | None:
        """
        Look up a single collection.

        :param lookup: A :ref:`lookup-single`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: The ID of the collection, or None if the provided lookup is
          None (for convenience in some call sites).
        :raises KeyError: if the lookup does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way.
        """
        # Import here to prevent circular imports
        from debusine.server.collections.lookup import lookup_single

        return (
            None
            if lookup is None
            else lookup_single(
                lookup=lookup,
                workspace=self.work_request.workspace,
                user=self.work_request.created_by,
                default_category=default_category,
                workflow_root=self.work_request.get_workflow_root(),
                expect_type=LookupChildType.COLLECTION,
            ).collection.id
        )

    def get_server_setting(self, setting: str) -> str:
        """Look up a Django setting (strings only)."""
        value = getattr(settings, setting)
        assert isinstance(value, str)
        return value



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the workspace models."""
import datetime
from typing import ClassVar
from unittest import mock

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import (
    ImproperlyConfigured,
    PermissionDenied,
    ValidationError,
)
from django.db import IntegrityError, transaction
from django.db.models import Q
from django.utils import timezone

from debusine.artifacts.models import CollectionCategory
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import (
    Collection,
    FileInArtifact,
    FileInStore,
    FileStore,
    Group,
    Scope,
    WorkRequest,
    Workspace,
    default_workspace,
)
from debusine.db.models.auth import User
from debusine.db.models.permissions import PartialCheckResult
from debusine.db.models.workspaces import (
    DeleteWorkspaces,
    WorkspaceChain,
    WorkspaceRole,
    WorkspaceRoleBase,
    WorkspaceRoles,
    is_valid_workspace_name,
)
from debusine.db.playground import scenarios
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    override_permission,
)


class WorkspaceRolesTests(TestCase):
    """Tests for the WorkspaceRoles class."""

    scenario = scenarios.DefaultContext()

    def test_init(self) -> None:
        r = WorkspaceRoleBase("test")
        r._setup()
        self.assertEqual(r.implied_by_scope_roles, frozenset())
        self.assertEqual(r.implied_by_workspace_roles, frozenset([r]))

    def test_init_invalid_implication(self) -> None:
        r = WorkspaceRoleBase("test", implied_by=[Q()])
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"Workspace roles do not support implications"
            r" by <class 'django.db.models.query_utils.Q'>",
        ):
            r._setup()

    def test_q_no_scope_implications(self) -> None:
        """Test generating a query with no scope implications."""
        user = self.scenario.user
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[user]
        )

        r = WorkspaceRoleBase("test")
        r._setup()
        self.assertQuerySetEqual(Workspace.objects.filter(r.q(user)), [])

        # Assign r as a role
        with context.disable_permission_checks():
            group = Group.objects.create(scope=self.scenario.scope, name="test")
            group.users.add(user)
            Workspace.objects.get_roles_model().objects.create(
                group=group, resource=self.scenario.workspace, role=r
            )
        self.assertQuerySetEqual(
            Workspace.objects.filter(r.q(user)), [self.scenario.workspace]
        )

    def test_choices(self) -> None:
        self.assertEqual(
            WorkspaceRoles.choices,
            [("owner", "Owner"), ("contributor", "Contributor")],
        )

    def test_implications(self) -> None:
        self.assertEqual(Workspace.Roles.OWNER.label, "Owner")
        self.assertEqual(
            Workspace.Roles.OWNER.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER]),
        )
        self.assertEqual(
            Workspace.Roles.OWNER.implied_by_workspace_roles,
            frozenset([Workspace.Roles.OWNER]),
        )

        self.assertEqual(Workspace.Roles.CONTRIBUTOR.label, "Contributor")
        self.assertEqual(
            Workspace.Roles.CONTRIBUTOR.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER]),
        )
        self.assertEqual(
            Workspace.Roles.CONTRIBUTOR.implied_by_workspace_roles,
            frozenset([Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR]),
        )


class WorkspaceManagerTests(TestCase):
    """Tests for the WorkspaceManager class."""

    scenario = scenarios.DefaultContext()
    scope1: ClassVar[Scope]

    @classmethod
    @context.disable_permission_checks()
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.scope1 = cls.playground.get_or_create_scope("Scope1")

    def test_get_roles_model(self) -> None:
        """Test the get_roles_model method."""
        self.assertIs(Workspace.objects.get_roles_model(), WorkspaceRole)

    def test_get_for_context_missing_scope(self) -> None:
        """Test get_for_context without scope in context."""
        context.reset()
        with self.assertRaisesRegex(
            ContextConsistencyError, r"scope is not set in context"
        ):
            Workspace.objects.get_for_context(self.scenario.workspace.name)

    def test_get_for_context_missing_user(self) -> None:
        """Test get_for_context without user in context."""
        context.reset()
        context.set_scope(self.scenario.scope)
        with self.assertRaisesRegex(
            ContextConsistencyError, r"user is not set in context"
        ):
            Workspace.objects.get_for_context(self.scenario.workspace.name)

    def test_get_for_context_notfound(self) -> None:
        """Test get_for_context using a wrong name."""
        self.scenario.set_current()
        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get_for_context("does-not-exist")

    def test_get_for_context_wrong_scope(self) -> None:
        """Test get_for_context with a name in the wrong scope."""
        context.reset()
        context.set_scope(self.scope1)
        context.set_user(self.scenario.user)
        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get_for_context(self.scenario.workspace.name)

    def test_get_for_context_anonymous_user(self) -> None:
        """Test get_for_context with an anonymous user."""
        context.reset()
        context.set_scope(self.scenario.scope)
        context.set_user(AnonymousUser())
        workspace = Workspace.objects.get_for_context(
            self.scenario.workspace.name
        )
        # Scope is prefetched
        with self.assertNumQueries(0):
            workspace.scope
        # Roles are empty
        self.assertFalse(hasattr(workspace, "user_roles"))

    def test_get_for_context_fetch_roles(self) -> None:
        """Ensure get_for_context fetches user roles."""
        self.scenario.set_current()
        # No roles
        workspace = Workspace.objects.get_for_context(
            self.scenario.workspace.name
        )
        self.assertEqual(getattr(workspace, "user_roles"), [])

        # Scope is prefetched
        with self.assertNumQueries(0):
            workspace.scope

        # Roles via one group
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[self.scenario.user],
        )
        workspace = Workspace.objects.get_for_context(
            self.scenario.workspace.name
        )
        self.assertEqual(
            getattr(workspace, "user_roles"), [Workspace.Roles.OWNER]
        )

        # Roles via multiple groups
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[self.scenario.user],
            name="more owners",
        )
        workspace = Workspace.objects.get_for_context(
            self.scenario.workspace.name
        )
        self.assertEqual(
            getattr(workspace, "user_roles"), [Workspace.Roles.OWNER]
        )

    def test_in_current_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter."""
        with context.local():
            self.scenario.set_current()
            self.assertQuerySetEqual(
                Workspace.objects.in_current_scope(), [self.scenario.workspace]
            )

        workspace1 = self.playground.create_workspace(scope=self.scope1)
        with context.local():
            context.set_scope(self.scope1)
            self.assertQuerySetEqual(
                Workspace.objects.in_current_scope(), [workspace1]
            )

    def test_in_current_scope_no_context_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter without scope set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope is not set"
        ):
            Workspace.objects.in_current_scope()


class WorkspaceTests(TestCase):
    """Tests for the Workspace class."""

    scenario = scenarios.DefaultContext()

    a: ClassVar[Workspace]
    b: ClassVar[Workspace]
    c: ClassVar[Workspace]
    d: ClassVar[Workspace]

    @staticmethod
    def _get_collection(
        workspace: Workspace,
        name: str,
        category: CollectionCategory = CollectionCategory.SUITE,
        user: User | AnonymousUser | None = None,
    ) -> Collection:
        """Shortcut to lookup a collection from a workspace."""
        if user is None:
            user = AnonymousUser()
        return workspace.get_collection(name=name, category=category, user=user)

    @classmethod
    @context.disable_permission_checks()
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.a = cls.playground.create_workspace(name="a", public=True)
        cls.b = cls.playground.create_workspace(name="b", public=True)
        cls.c = cls.playground.create_workspace(name="c", public=True)
        cls.d = cls.playground.create_workspace(name="d", public=True)

    def assertInherits(
        self, child: Workspace, parents: list[Workspace]
    ) -> None:
        """Ensure inheritance chain matches."""
        chain = list(
            child.chain_parents.order_by("order").values_list(
                "child__name", "parent__name", "order"
            )
        )

        expected: list[tuple[str, str, int]] = []
        for idx, parent in enumerate(parents):
            expected.append((child.name, parent.name, idx))
        self.assertEqual(chain, expected)

    def assertGetCollectionEqual(
        self,
        workspace: Workspace,
        name: str,
        expected: Collection,
        *,
        category: CollectionCategory = CollectionCategory.SUITE,
        user: User | None = None,
    ) -> None:
        """Check that collection lookup yields the given result."""
        self.assertEqual(
            self._get_collection(
                workspace, name=name, category=category, user=user
            ),
            expected,
        )

    def assertGetCollectionFails(
        self,
        workspace: Workspace,
        name: str,
        *,
        category: CollectionCategory = CollectionCategory.SUITE,
        user: User | None = None,
    ) -> None:
        """Check that collection lookup fails."""
        with self.assertRaises(Collection.DoesNotExist):
            self._get_collection(
                workspace, name=name, category=category, user=user
            )

    def assertPass(self, value: PartialCheckResult) -> None:
        """Check that the value is PASS."""
        self.assertEqual(value, PartialCheckResult.PASS)

    def test_get_absolute_url(self) -> None:
        """Test the get_absolute_url method."""
        self.assertEqual(
            self.scenario.workspace.get_absolute_url(),
            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/System/",
        )

    def test_get_absolute_url_configure(self) -> None:
        """Test the get_absolute_url_configure method."""
        self.assertEqual(
            self.scenario.workspace.get_absolute_url_configure(),
            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/System/configure/",
        )

    def test_is_valid_workspace_name(self) -> None:
        """Test is_valid_workspace_name."""
        valid_names = (
            "debian",
            "debusine",
            "System",
            "c++",
            "foo_bar",
            "foo.bar",
            "foo+bar",
            "tail_",
            "c--",
        )
        invalid_names = (
            "artifact",
            "workspaces",
            "+tag",
            "-tag",
            ".profile",
            "_reserved",
            "foo:bar",
        )

        for name in valid_names:
            with self.subTest(name=name):
                self.assertTrue(is_valid_workspace_name(name))

        for name in invalid_names:
            with self.subTest(name=name):
                self.assertFalse(is_valid_workspace_name(name))

    def test_workspace_name_validation(self) -> None:
        """Test validation for workspace names."""
        kwargs = {"scope": self.scenario.scope}
        Workspace(name="foo", **kwargs).full_clean()
        Workspace(name="foo_", **kwargs).full_clean()
        with self.assertRaises(ValidationError) as exc:
            Workspace(name="_foo", **kwargs).full_clean()
        self.assertEqual(
            exc.exception.message_dict,
            {'name': ["'_foo' is not a valid workspace name"]},
        )

    @context.disable_permission_checks()
    def test_default_values_fields(self) -> None:
        """Test basic behavior."""
        name = "test"
        workspace = Workspace(name=name, scope=self.scenario.scope)

        self.assertEqual(workspace.name, name)
        self.assertFalse(workspace.public)
        self.assertIsNone(workspace.created_at)
        self.assertIsNone(workspace.expiration_delay)

        pre_save_time = timezone.now()

        workspace.clean_fields()
        workspace.save()
        self.assertGreaterEqual(workspace.created_at, pre_save_time)

    @context.disable_permission_checks()
    def test_scoping(self) -> None:
        """Test scoping."""
        scope1 = self.playground.get_or_create_scope("scope1")
        scope2 = self.playground.get_or_create_scope("scope2")
        Workspace.objects.create(name="test", scope=scope1)
        Workspace.objects.create(name="test", scope=scope2)
        with (
            transaction.atomic(),
            self.assertRaisesRegex(
                IntegrityError,
                # Only match constraint name to support non-english locales
                # "duplicate key value violates unique constraint"
                "db_workspace_unique_scope_name",
            ),
        ):
            Workspace.objects.create(name="test", scope=scope1)

    def test_context_has_role_no_context(self) -> None:
        """Test context_has_role with no context set."""
        self.assertPass(
            self.scenario.workspace.context_has_role(
                self.scenario.user, Workspace.Roles.OWNER
            )
        )

    def test_context_has_role_wrong_user(self) -> None:
        """Test context_has_role with a different user than in context."""
        self.scenario.set_current()
        user = self.playground.create_user("test")
        self.assertPass(
            self.scenario.workspace.context_has_role(
                user, Workspace.Roles.OWNER
            ),
        )

    def test_context_has_role_wrong_scope(self) -> None:
        """Test context_has_role with a different scope than in context."""
        self.scenario.set_current()
        scope = self.playground.get_or_create_scope("test")
        workspace = self.playground.create_workspace(name="test", scope=scope)
        self.assertPass(
            workspace.context_has_role(
                self.scenario.user, Workspace.Roles.OWNER
            )
        )

    def test_context_has_role_wrong_workspace(self) -> None:
        """Test context_has_role with a different workspace than in context."""
        # User is owner of the scope, not of the workspace
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[self.scenario.user]
        )

        self.scenario.set_current()
        workspace = self.playground.create_workspace(name="test")

        # Not enough information to decide about workspace roles
        self.assertPass(
            workspace.context_has_role(
                self.scenario.user, Workspace.Roles.OWNER
            )
        )

    def test_context_has_role(self) -> None:
        """Test context_has_role."""
        for scope_roles, workspace_roles, roles, expected in (
            (
                [],
                [],
                [Workspace.Roles.CONTRIBUTOR, Workspace.Roles.OWNER],
                PartialCheckResult.DENY,
            ),
            (
                [Scope.Roles.OWNER],
                [],
                [Workspace.Roles.CONTRIBUTOR, Workspace.Roles.OWNER],
                PartialCheckResult.ALLOW,
            ),
            (
                [],
                [Workspace.Roles.OWNER],
                [Workspace.Roles.CONTRIBUTOR, Workspace.Roles.OWNER],
                PartialCheckResult.ALLOW,
            ),
            (
                [],
                [Workspace.Roles.CONTRIBUTOR],
                [Workspace.Roles.CONTRIBUTOR],
                PartialCheckResult.ALLOW,
            ),
            (
                [],
                [Workspace.Roles.CONTRIBUTOR],
                [Workspace.Roles.OWNER],
                PartialCheckResult.DENY,
            ),
        ):
            for role in roles:
                with (
                    self.subTest(
                        scope_roles=scope_roles,
                        workspace_roles=workspace_roles,
                        role=role,
                    ),
                    context.local(),
                ):
                    Group.objects.all().delete()
                    for sr in scope_roles:
                        self.playground.create_group_role(
                            self.scenario.scope, sr, users=[self.scenario.user]
                        )
                    for wr in workspace_roles:
                        self.playground.create_group_role(
                            self.scenario.workspace,
                            wr,
                            users=[self.scenario.user],
                        )
                    self.scenario.set_current()
                    self.assertEqual(
                        self.scenario.workspace.context_has_role(
                            self.scenario.user, role
                        ),
                        expected,
                    )

    @context.disable_permission_checks()
    def test_file_needs_upload_initial_file_store(self) -> None:
        """
        Test file_needs_upload returns the correct values.

        The file is added in the scope's initial FileStore.
        """
        fileobj = self.playground.create_file()
        workspace = default_workspace()
        scope = workspace.scope
        initial_file_store = scope.file_stores.get()
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add file in store
        FileInStore.objects.create(file=fileobj, store=initial_file_store)

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [initial_file_store]
        )
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add file to an artifact in another workspace
        other_workspace = self.playground.create_workspace(name="other")
        other_artifact, _ = self.playground.create_artifact(
            workspace=other_workspace
        )
        FileInArtifact.objects.create(
            artifact=other_artifact, path="test", file=fileobj
        )

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [initial_file_store]
        )
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add different but complete file to an artifact in this workspace
        artifact, _ = self.playground.create_artifact(workspace=workspace)
        FileInArtifact.objects.create(
            artifact=artifact,
            path="other",
            file=self.playground.create_file(b"other"),
            complete=True,
        )

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [initial_file_store]
        )
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add incomplete file to an artifact in this workspace
        file_in_artifact = FileInArtifact.objects.create(
            artifact=artifact, path="test", file=fileobj, complete=False
        )

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [initial_file_store]
        )
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Mark the previous file as complete
        file_in_artifact.complete = True
        file_in_artifact.save()

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [initial_file_store]
        )
        self.assertFalse(workspace.file_needs_upload(fileobj))

    @context.disable_permission_checks()
    def test_file_needs_upload_other_file_store(self) -> None:
        """
        Test file_stores and file_needs_upload return the correct values.

        The file is added in a FileStore other than the scope's initial one.
        """
        fileobj = self.playground.create_file()
        workspace = default_workspace()
        scope = workspace.scope
        self.assertQuerySetEqual(scope.download_file_stores(fileobj), [])
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add file in a store which is not the initial one
        store = FileStore.objects.create(
            name="nas-01", backend=FileStore.BackendChoices.LOCAL
        )
        scope.file_stores.add(store)
        workspace.refresh_from_db()

        self.assertQuerySetEqual(scope.download_file_stores(fileobj), [])
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add file in the store
        FileInStore.objects.create(file=fileobj, store=store)

        self.assertQuerySetEqual(scope.download_file_stores(fileobj), [store])
        self.assertTrue(workspace.file_needs_upload(fileobj))

        # Add file to an artifact in this workspace
        artifact, _ = self.playground.create_artifact(workspace=workspace)
        FileInArtifact.objects.create(
            artifact=artifact, path="test", file=fileobj, complete=True
        )

        self.assertQuerySetEqual(scope.download_file_stores(fileobj), [store])
        self.assertFalse(workspace.file_needs_upload(fileobj))

        # Add file to the initial store as well
        FileInStore.objects.create(
            file=fileobj, store=scope.file_stores.earliest("id")
        )

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj),
            [scope.file_stores.earliest("id"), store],
        )
        self.assertFalse(workspace.file_needs_upload(fileobj))

    def test_str(self) -> None:
        """Test __str__ method."""
        workspace = Workspace(name="test", scope=self.scenario.scope)

        self.assertEqual(workspace.__str__(), "debusine/test")

    def test_workspacechain_str(self) -> None:
        """Test WorkspaceChain.__str__ method."""
        chain = WorkspaceChain(parent=self.a, child=self.b, order=5)
        self.assertEqual(chain.__str__(), "5:b→a")

    def test_set_inheritance(self) -> None:
        """Test set_inheritance method."""
        a, b, c = self.a, self.b, self.c

        a.set_inheritance([b])
        self.assertInherits(a, [b])
        self.assertInherits(b, [])
        self.assertInherits(c, [])

        a.set_inheritance([b, c])
        self.assertInherits(a, [b, c])
        self.assertInherits(b, [])
        self.assertInherits(c, [])

        a.set_inheritance([c, b])
        self.assertInherits(a, [c, b])
        self.assertInherits(b, [])
        self.assertInherits(c, [])

        c.set_inheritance([a])
        self.assertInherits(a, [c, b])
        self.assertInherits(b, [])
        self.assertInherits(c, [a])

        a.set_inheritance([])
        self.assertInherits(a, [])
        self.assertInherits(b, [])
        self.assertInherits(c, [a])

        with self.assertRaisesRegex(
            ValueError, r"duplicate workspace 'b' in inheritance chain"
        ):
            a.set_inheritance([b, b, c])

    def test_collection_lookup(self) -> None:
        """Test collection lookup."""
        a, b, c = self.a, self.b, self.c

        a1 = self.playground.create_collection(
            name="a", category=CollectionCategory.SUITE, workspace=a
        )

        # Lookup locally
        self.assertEqual(self._get_collection(a, "a"), a1)
        self.assertGetCollectionFails(b, "a")
        self.assertGetCollectionFails(c, "a")

        # Lookup through a simple chain
        b.set_inheritance([a])
        self.assertGetCollectionEqual(b, "a", a1)

        # Walk the inheritance chain until the end
        b.set_inheritance([c, a])
        self.assertGetCollectionEqual(b, "a", a1)
        self.assertGetCollectionFails(c, "a")

        # Create a chain loop, lookup does not break
        c.set_inheritance([b])
        self.assertGetCollectionEqual(b, "a", a1)

    def test_collection_lookup_graph(self) -> None:
        """Test collection lookup graph."""
        a, b, c, d = self.a, self.b, self.c, self.d

        cb = self.playground.create_collection(
            name="a", category=CollectionCategory.SUITE, workspace=b
        )
        cc = self.playground.create_collection(
            name="a", category=CollectionCategory.SUITE, workspace=c
        )

        # Lookup happens in order
        a.set_inheritance([b, c])
        self.assertGetCollectionEqual(a, "a", cb)
        a.set_inheritance([c, b])
        self.assertGetCollectionEqual(a, "a", cc)

        # Lookup happens depth-first
        a.set_inheritance([d, c])
        b.set_inheritance([])
        c.set_inheritance([])
        d.set_inheritance([b])
        self.assertGetCollectionEqual(a, "a", cb)

    def test_user_restrictions(self) -> None:
        """Test user restriction enforcement."""
        user = self.scenario.user
        with context.disable_permission_checks():
            wpub = self.playground.create_workspace(name="public", public=True)
            cpub = self.playground.create_collection(
                "test", CollectionCategory.SUITE, workspace=wpub
            )
            wpriv = self.playground.create_workspace(
                name="private", public=False
            )
            cpriv = self.playground.create_collection(
                "test", CollectionCategory.SUITE, workspace=wpriv
            )
            wstart = self.playground.create_workspace(name="start", public=True)

            self.playground.create_group_role(
                wpriv, Workspace.Roles.OWNER, users=[user]
            )

            other_user = User.objects.create_user(
                username="other", email="other@example.org"
            )

        self.assertGetCollectionFails(wstart, "test", user=None)
        self.assertGetCollectionFails(wstart, "test", user=user)
        self.assertGetCollectionFails(wstart, "test", user=other_user)

        # Lookups in the workspace itself do check restrictions
        self.assertGetCollectionEqual(wpub, "test", cpub, user=None)
        self.assertGetCollectionEqual(wpub, "test", cpub, user=user)
        self.assertGetCollectionEqual(wpub, "test", cpub, user=other_user)
        self.assertGetCollectionFails(wpriv, "test", user=None)
        self.assertGetCollectionEqual(wpriv, "test", cpriv, user=user)
        self.assertGetCollectionFails(wpriv, "test", user=other_user)

        # Inheritance chain is always followed for public datasets
        wstart.set_inheritance([wpub])
        self.assertGetCollectionEqual(wstart, "test", cpub, user=None)
        self.assertGetCollectionEqual(wstart, "test", cpub, user=user)
        self.assertGetCollectionEqual(wstart, "test", cpub, user=other_user)

        # Inheritance chain on private datasets is followed only if logged in
        wstart.set_inheritance([wpriv])
        self.assertGetCollectionFails(wstart, "test", user=None)
        self.assertGetCollectionEqual(wstart, "test", cpriv, user=user)
        self.assertGetCollectionFails(wstart, "test", user=other_user)

        # Inheritance chain skips private datasets but can see public ones
        wstart.set_inheritance([wpriv, wpub])
        self.assertGetCollectionEqual(wstart, "test", cpub, user=None)
        self.assertGetCollectionEqual(wstart, "test", cpriv, user=user)
        self.assertGetCollectionEqual(wstart, "test", cpub, user=other_user)

    def test_get_roles_not_authenticated(self) -> None:
        """Test Workspace.get_roles when not authenticated."""
        user = AnonymousUser()
        self.assertQuerySetEqual(self.a.get_roles(user), [])

    def test_get_roles(self) -> None:
        """Test Workspace.get_roles."""
        group = self.playground.create_group_role(self.a, Workspace.Roles.OWNER)
        user = self.playground.get_default_user()
        self.assertQuerySetEqual(self.a.get_roles(user), [])

        self.playground.add_user(group, user)
        self.assertEqual(list(self.a.get_roles(user)), [Workspace.Roles.OWNER])

    def test_get_roles_multiple_groups(self) -> None:
        """Test Workspace.get_roles with the user in multiple groups."""
        group1 = self.playground.create_group_role(
            self.a, Workspace.Roles.OWNER, name="Group1"
        )
        group2 = self.playground.create_group_role(
            self.a, Workspace.Roles.OWNER, name="Group2"
        )

        user = self.playground.get_default_user()
        self.assertQuerySetEqual(self.a.get_roles(user), [])

        self.playground.add_user(group1, user)
        self.assertEqual(list(self.a.get_roles(user)), [Workspace.Roles.OWNER])

        self.playground.add_user(group2, user)
        self.assertEqual(list(self.a.get_roles(user)), [Workspace.Roles.OWNER])

    def test_set_current_without_scope(self) -> None:
        """Test set_current without scope."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "Cannot set workspace before scope"
        ):
            self.a.set_current()
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())

    def test_set_current_without_user(self) -> None:
        """Test set_current without user."""
        context.set_scope(self.scenario.scope)
        with self.assertRaisesRegex(
            ContextConsistencyError, "Cannot set workspace before user"
        ):
            self.a.set_current()
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())

    def test_set_current_with_token(self) -> None:
        """Test set_current without user."""
        context.set_scope(self.scenario.scope)
        context.set_worker_token(self.playground.create_worker_token())
        self.a.set_current()
        self.assertEqual(context.workspace, self.a)
        self.assertEqual(context.workspace_roles, frozenset())

    def test_context_cannot_change_workspace(self) -> None:
        """Changing workspace in context is not allowed."""
        self.scenario.set_current()

        with self.assertRaisesRegex(
            ContextConsistencyError,
            "Workspace was already set to debusine/System",
        ):
            self.b.set_current()

        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.user, self.scenario.user)
        self.assertEqual(context.workspace, self.scenario.workspace)

    def test_set_current_wrong_scope(self) -> None:
        """Test scope/workspace.scope consistency checks."""
        scope1 = Scope.objects.create(name="scope1")
        self.scenario.workspace.scope = scope1
        self.scenario.workspace.save()

        context.set_scope(self.scenario.scope)
        context.set_user(self.scenario.user)
        with self.assertRaisesRegex(
            ContextConsistencyError,
            "workspace scope 'scope1' does not match current scope 'debusine'",
        ):
            self.scenario.workspace.set_current()

        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.user, self.scenario.user)
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())

    def test_set_current_forbidden_workspace(self) -> None:
        """Test set_current with an inaccessible workspace."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        context.set_scope(self.scenario.scope)
        context.set_user(self.scenario.user)
        with self.assertRaisesRegex(
            ContextConsistencyError,
            "User playground cannot access workspace debusine/System",
        ):
            self.scenario.workspace.set_current()

        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.user, self.scenario.user)
        self.assertIsNone(context.workspace)
        self.assertIsNone(context._workspace_roles.get())

    def test_set_current(self) -> None:
        """Test set_current."""
        context.set_scope(self.scenario.scope)
        context.set_user(self.scenario.user)
        # One query to fetch the list of workspace roles
        with self.assertNumQueries(1):
            self.a.set_current()
        self.assertEqual(context.workspace_roles, frozenset())

    def test_set_current_cached_roles(self) -> None:
        """Test set_current with user roles cached in Workspace."""
        workspace = self.scenario.workspace
        workspace.public = False
        workspace.save()
        context.set_scope(self.scenario.scope)
        context.set_user(self.scenario.user)
        for role in Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR:
            with self.subTest(role=role), context.local():
                setattr(workspace, "user_roles", [role])
                with self.assertNumQueries(0):
                    workspace.set_current()
                self.assertEqual(context.workspace_roles, frozenset((role,)))

    def test_predicate_deny_from_context(self) -> None:
        """Test predicates propagating DENY from context_has_role."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        with mock.patch(
            "debusine.db.models.workspaces.Workspace.context_has_role",
            return_value=PartialCheckResult.DENY,
        ):
            self.assertFalse(
                self.scenario.workspace.can_display(self.scenario.user)
            )
            self.assertFalse(
                self.scenario.workspace.can_configure(self.scenario.user)
            )
            self.assertFalse(
                self.scenario.workspace.can_create_artifacts(self.scenario.user)
            )
            self.assertFalse(
                self.scenario.workspace.can_create_experiment_workspace(
                    self.scenario.user
                )
            )

    def test_can_display_public(self) -> None:
        """Test the can_display predicate on public workspaces."""
        self.assertPermission(
            "can_display",
            users=(AnonymousUser(), self.scenario.user),
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

    def test_can_display_current(self) -> None:
        """Test the can_display predicate on the current workspace."""
        user1 = self.playground.create_user("test1")
        self.a.public = False
        self.a.save()
        for role in Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR:
            with self.subTest(role=role), context.local():
                self.playground.create_group_role(
                    self.a, role, users=[self.scenario.user, user1]
                )

                context.set_scope(self.scenario.scope)
                context.set_user(self.scenario.user)
                self.a.set_current()

                # can_display with current workspace and user is shortcut
                with self.assertNumQueries(0):
                    self.assertTrue(self.a.can_display(self.scenario.user))

                with self.assertNumQueries(1):
                    self.assertTrue(self.a.can_display(user1))

    def test_can_display_by_roles(self) -> None:
        """Test can_display workspace behaviour with roles."""
        # Make workspace private
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        self.assertPermissionWhenRole(
            self.scenario.workspace.can_display,
            self.scenario.user,
            (Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR),
            scope_roles=Scope.Roles.OWNER,
        )

    def test_can_display_with_token(self) -> None:
        """Test can_display with a worker token."""
        # Make a private
        self.a.public = False
        self.a.save()

        # Workspace is not accessible
        self.assertPermission(
            "can_display",
            users=(AnonymousUser(), self.scenario.user),
            allowed=[self.b, self.c, self.d, self.scenario.workspace],
            denied=self.a,
        )

        # Workspace is now accessible also without user information
        self.assertPermission(
            "can_display",
            users=(None, AnonymousUser(), self.scenario.user),
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
            token=self.playground.create_worker_token(),
        )

    def test_can_configure_public(self) -> None:
        """Test the can_configure predicate on public workspaces."""
        self.assertPermission(
            "can_configure",
            users=(AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

    def test_can_configure_current(self) -> None:
        """Test the can_configure predicate on the current workspace."""
        user1 = self.playground.create_user("test1")
        for role, expected in (
            (Workspace.Roles.OWNER, True),
            (Workspace.Roles.CONTRIBUTOR, False),
        ):
            with self.subTest(role=role), context.local():
                group = self.playground.create_group_role(
                    self.a, role, users=[self.scenario.user, user1]
                )

                context.set_scope(self.scenario.scope)
                context.set_user(self.scenario.user)
                self.a.set_current()

                # can_configure with current workspace and user is shortcut
                with self.assertNumQueries(0):
                    self.assertEqual(
                        self.a.can_configure(self.scenario.user), expected
                    )

                with self.assertNumQueries(1):
                    self.assertEqual(self.a.can_configure(user1), expected)

                group.delete()

    def test_can_configure_by_roles(self) -> None:
        """Test can_configure workspace behaviour with roles."""
        self.assertPermissionWhenRole(
            self.scenario.workspace.can_configure,
            self.scenario.user,
            Workspace.Roles.OWNER,
            scope_roles=Scope.Roles.OWNER,
        )

    def test_can_configure_with_token(self) -> None:
        """Test can_configure with a worker token."""
        # Workspace is not configurable
        self.assertPermission(
            "can_configure",
            users=(AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

        # Workspace is also not configurable with a worker token
        self.assertPermission(
            "can_configure",
            users=(None, AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
            token=self.playground.create_worker_token(),
        )

    def test_create_checks(self) -> None:
        """Test can_create_workspace hook in save."""
        with mock.patch(
            "debusine.db.models.Scope.can_create_workspace"
        ) as pred:
            test1 = Workspace.objects.create(
                name="test1", scope=self.scenario.scope
            )
        pred.assert_called_once_with(None)
        self.assertIsNotNone(test1.pk)

        with override_permission(Scope, "can_create_workspace", AllowAll):
            test2 = Workspace.objects.create(
                name="test2", scope=self.scenario.scope
            )
        self.assertIsNotNone(test2.pk)

        with override_permission(Scope, "can_create_workspace", DenyAll):
            with self.assertRaisesRegex(
                PermissionDenied, r"None cannot create workspaces in debusine"
            ):
                Workspace.objects.create(
                    name="test3", scope=self.scenario.scope
                )
        self.assertQuerySetEqual(Workspace.objects.filter(name="test3"), [])

    def test_can_create_artifacts_public(self) -> None:
        """Test the can_create_artifacts predicate on public workspaces."""
        self.assertPermission(
            "can_create_artifacts",
            users=AnonymousUser(),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )
        self.assertPermission(
            "can_create_artifacts",
            users=self.scenario.user,
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

    def test_can_create_artifacts_current(self) -> None:
        """Test the can_create_artifacts predicate on the current workspace."""
        user1 = self.playground.create_user("test1")
        self.a.public = False
        self.a.save()
        for role in Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR:
            with self.subTest(role=role), context.local():
                self.playground.create_group_role(
                    self.a, role, users=[self.scenario.user, user1]
                )

                context.set_scope(self.scenario.scope)
                context.set_user(self.scenario.user)
                self.a.set_current()

                # can_create_artifacts with current workspace and user is
                # shortcut
                with self.assertNumQueries(0):
                    self.assertTrue(
                        self.a.can_create_artifacts(self.scenario.user)
                    )

                with self.assertNumQueries(1):
                    self.assertTrue(self.a.can_create_artifacts(user1))

    def test_can_create_artifacts_by_roles(self) -> None:
        """Test can_create_artifacts behaviour with roles."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        self.assertPermissionWhenRole(
            self.scenario.workspace.can_create_artifacts,
            self.scenario.user,
            [Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR],
            scope_roles=Scope.Roles.OWNER,
        )

    def test_can_create_artifacts_with_token(self) -> None:
        """Test can_create_artifacts with a worker token."""
        # Make a private
        self.a.public = False
        self.a.save()

        # Workspace is not accessible
        self.assertPermission(
            "can_create_artifacts",
            users=AnonymousUser(),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )
        self.assertPermission(
            "can_create_artifacts",
            users=self.scenario.user,
            allowed=[self.b, self.c, self.d, self.scenario.workspace],
            denied=self.a,
        )

        # Workspace is now accessible also without user information
        self.assertPermission(
            "can_create_artifacts",
            users=(None, AnonymousUser(), self.scenario.user),
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
            token=self.playground.create_worker_token(),
        )

    def test_can_create_work_requests_public(self) -> None:
        """Test can_create_work_requests on public workspaces."""
        self.assertPermission(
            "can_create_work_requests",
            users=(AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

    def test_can_create_work_requests_current_owner(self) -> None:
        """Test can_create_work_requests on current owner."""
        user1 = self.playground.create_user("test1")
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[self.scenario.user, user1],
        )

        self.scenario.set_current()

        # Test shortcut code path
        with self.assertNumQueries(0):
            self.assertTrue(
                self.scenario.workspace.can_create_work_requests(
                    self.scenario.user
                )
            )

        with self.assertNumQueries(1):
            self.assertTrue(
                self.scenario.workspace.can_create_work_requests(user1)
            )

    def test_can_create_work_requests_current_contributor(self) -> None:
        """Test can_create_work_requests on current contributor."""
        user1 = self.playground.create_user("test1")
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.scenario.user, user1],
        )

        self.scenario.set_current()

        # Test shortcut code path
        with self.assertNumQueries(0):
            self.assertTrue(
                self.scenario.workspace.can_create_work_requests(
                    self.scenario.user
                )
            )

        with self.assertNumQueries(1):
            self.assertTrue(
                self.scenario.workspace.can_create_work_requests(user1)
            )

    def test_can_create_work_requests_by_roles(self) -> None:
        """Test can_create_work_requests behaviour with roles."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        self.assertPermissionWhenRole(
            self.scenario.workspace.can_create_work_requests,
            self.scenario.user,
            (Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR),
            scope_roles=Scope.Roles.OWNER,
        )

    def test_can_create_work_requests_with_token(self) -> None:
        """Test can_create_work_requests with a worker token."""
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[self.scenario.user]
        )

        # Anonymous cannot
        self.assertPermission(
            "can_create_work_requests",
            users=AnonymousUser(),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

        # User can
        self.assertPermission(
            "can_create_work_requests",
            users=self.scenario.user,
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

        # A worker token cannot in any case
        self.assertPermission(
            "can_create_work_requests",
            users=(None, AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
            token=self.playground.create_worker_token(),
        )

    def test_can_create_experiment_workspace_public(self) -> None:
        """Test can_create_experiment_workspace on public workspaces."""
        self.assertPermission(
            "can_create_experiment_workspace",
            users=AnonymousUser(),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )
        self.assertPermission(
            "can_create_experiment_workspace",
            users=self.scenario.user,
            allowed=[self.a, self.b, self.c, self.d, self.scenario.workspace],
        )

    def test_can_create_experiment_workspace_current(self) -> None:
        """Test can_create_experiment_workspace on the current workspace."""
        user1 = self.playground.create_user("test1")
        self.a.public = False
        self.a.save()
        for role in Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR:
            with self.subTest(role=role), context.local():
                self.playground.create_group_role(
                    self.a, role, users=[self.scenario.user, user1]
                )

                context.set_scope(self.scenario.scope)
                context.set_user(self.scenario.user)
                self.a.set_current()

                # can_create_experiment_workspace with current workspace and
                # user is shortcut
                with self.assertNumQueries(0):
                    self.assertTrue(
                        self.a.can_create_experiment_workspace(
                            self.scenario.user
                        )
                    )

                with self.assertNumQueries(1):
                    self.assertTrue(
                        self.a.can_create_experiment_workspace(user1)
                    )

    def test_can_create_experiment_workspace_by_roles(self) -> None:
        """Test can_create_experiment_workspace workspace with roles."""
        # Make workspace private
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        self.assertPermissionWhenRole(
            self.scenario.workspace.can_create_experiment_workspace,
            self.scenario.user,
            (Workspace.Roles.OWNER, Workspace.Roles.CONTRIBUTOR),
            scope_roles=Scope.Roles.OWNER,
        )

    def test_can_create_experiment_workspace_with_token(self) -> None:
        """Test can_create_experiment_workspace with a worker token."""
        # Make a private
        self.a.public = False
        self.a.save()

        # Using a worker token, the permission is always denied
        self.assertPermission(
            "can_create_experiment_workspace",
            users=(None, AnonymousUser(), self.scenario.user),
            denied=[self.a, self.b, self.c, self.d, self.scenario.workspace],
            token=self.playground.create_worker_token(),
        )

    def test_with_expiration_time_no_expiry(self) -> None:
        """Test with_expiration_time with no expiry."""
        ws = Workspace.objects.with_expiration_time().get(name="System")
        self.assertIsNone(ws.expiration_time)

    def test_with_expiration_time_created_at(self) -> None:
        """Test with_expiration_time with only created_at."""
        self.a.expiration_delay = datetime.timedelta(days=3)
        self.a.save()
        ws = Workspace.objects.with_expiration_time().get(pk=self.a.pk)
        self.assertEqual(
            ws.expiration_time, self.a.created_at + self.a.expiration_delay
        )

    def test_with_expiration_time_work_request_completed(self) -> None:
        """Test with_expiration_time with only created_at."""
        self.a.expiration_delay = datetime.timedelta(days=3)
        self.a.save()
        wr = self.playground.create_work_request(
            result=WorkRequest.Results.SUCCESS, workspace=self.a
        )
        wr.completed_at = self.a.created_at + datetime.timedelta(days=42)
        wr.save()
        ws = Workspace.objects.with_expiration_time().get(pk=self.a.pk)
        self.assertEqual(
            ws.expiration_time,
            self.a.created_at
            + self.a.expiration_delay
            + datetime.timedelta(days=42),
        )

    def test_expire_at_no_expiry(self) -> None:
        """Test expire_at with no expiry."""
        self.assertIsNone(self.scenario.workspace.expire_at)

    def test_expire_at(self) -> None:
        """Test with_expiration_time, computed by expire_at."""
        self.a.expiration_delay = datetime.timedelta(days=3)
        self.a.save()
        self.assertFalse(hasattr(self.a, "expiration_time"))
        expiration_time = self.a.created_at + self.a.expiration_delay
        self.assertEqual(self.a.expire_at, expiration_time)
        self.assertEqual(
            self.a.expiration_time,  # type: ignore[attr-defined]
            expiration_time,
        )

    def test_expire_at_precomputed(self) -> None:
        """Test with_expiration_time precomputed by with_expiration_time."""
        self.a.expiration_delay = datetime.timedelta(days=3)
        self.a.save()
        expiration_time = self.a.created_at + self.a.expiration_delay
        a = Workspace.objects.with_expiration_time().get(pk=self.a.pk)
        self.assertEqual(
            a.expiration_time,
            expiration_time,
        )
        self.assertEqual(a.expire_at, expiration_time)


class WorkspaceRoleTests(TestCase):
    """Tests for the WorkspaceRole class."""

    workspace1: ClassVar[Workspace]
    workspace2: ClassVar[Workspace]
    group1: ClassVar[Group]
    group2: ClassVar[Group]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common test data."""
        super().setUpTestData()
        with context.disable_permission_checks():
            cls.workspace1 = cls.playground.create_workspace(name="Workspace1")
            cls.workspace2 = cls.playground.create_workspace(name="Workspace2")
            cls.group1 = cls.playground.create_group("Group1")
            cls.group2 = cls.playground.create_group("Group2")

    def test_str(self) -> None:
        """Test stringification."""
        sr = WorkspaceRole(
            group=self.group1,
            resource=self.workspace1,
            role=WorkspaceRole.Roles.OWNER,
        )
        self.assertEqual(str(sr), "debusine/Group1─owner⟶debusine/Workspace1")

    def test_assign_multiple_groups(self) -> None:
        """Multiple group can share a role on a workspace."""
        WorkspaceRole.objects.create(
            group=self.group1,
            resource=self.workspace1,
            role=WorkspaceRole.Roles.OWNER,
        )
        WorkspaceRole.objects.create(
            group=self.group2,
            resource=self.workspace1,
            role=WorkspaceRole.Roles.OWNER,
        )

    def test_assign_multiple_scopes(self) -> None:
        """A group can be given a role in different workspaces."""
        WorkspaceRole.objects.create(
            group=self.group1,
            resource=self.workspace1,
            role=WorkspaceRole.Roles.OWNER,
        )
        WorkspaceRole.objects.create(
            group=self.group1,
            resource=self.workspace2,
            role=WorkspaceRole.Roles.OWNER,
        )

    def test_on_delete_cascade(self) -> None:
        """Test on delete cascade behaviour."""
        assignment = WorkspaceRole.objects.create(
            group=self.group1,
            resource=self.workspace1,
            role=WorkspaceRole.Roles.OWNER,
        )
        self.group1.delete()
        self.assertFalse(
            WorkspaceRole.objects.filter(pk=assignment.pk).exists()
        )

        assignment = WorkspaceRole.objects.create(
            group=self.group2,
            resource=self.workspace2,
            role=WorkspaceRole.Roles.OWNER,
        )
        self.workspace2.delete()
        self.assertFalse(
            WorkspaceRole.objects.filter(pk=assignment.pk).exists()
        )


class WorkspaceDeleteTests(TestCase):
    """Test DeleteWorkspaces."""

    scenario = scenarios.UIPlayground()

    def test_scan(self) -> None:
        """Test DeleteWorkspaces scan of resources."""
        operation = DeleteWorkspaces(
            Workspace.objects.filter(pk=self.scenario.workspace.pk)
        )

        self.assertQuerySetEqual(
            operation.workflow_templates, [self.scenario.template_sbuild]
        )
        self.assertEqual(operation.artifacts.count(), 17)
        self.assertEqual(operation.work_requests.count(), 12)
        self.assertEqual(operation.collection_items.count(), 7)
        self.assertEqual(operation.collections.count(), 6)
        self.assertEqual(operation.file_in_artifacts.count(), 18)
        self.assertEqual(operation.artifact_relations.count(), 14)
        self.assertEqual(operation.asset_usages.count(), 1)
        self.assertEqual(operation.assets.count(), 1)
        self.assertEqual(operation.workspace_roles.count(), 1)

    def test_delete(self) -> None:
        """Test DeleteWorkspaces deletion."""
        operation = DeleteWorkspaces(
            Workspace.objects.filter(pk=self.scenario.workspace.pk)
        )
        operation.perform_deletions()
        self.assertFalse(
            Workspace.objects.filter(pk=self.scenario.workspace.pk).exists()
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_file_store."""

import io

import yaml
from django.test import override_settings

from debusine.db.models import FileStore
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


@override_settings(LANGUAGE_CODE="en-us")
class CreateFileStoreCommandTests(TestCase):
    """Tests for the create_file_store command."""

    def test_forwards_with_deprecation_warning(self) -> None:
        """`create_file_store` forwards to `file_store create` and warns."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        with self.assertWarnsMessage(
            DeprecationWarning,
            "The `debusine-admin create_file_store` command has been "
            "deprecated in favour of `debusine-admin file_store create`",
        ):
            stdout, stderr, exit_code = call_command(
                "create_file_store",
                name,
                backend,
                stdin=io.StringIO(yaml.safe_dump(configuration)),
            )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data migration for FileInArtifact.complete."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps


def update_file_in_artifact(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Populate `FileInArtifact.complete`."""
    FileInArtifact = apps.get_model("db", "FileInArtifact")
    FileStore = apps.get_model("db", "FileStore")
    Workspace = apps.get_model("db", "Workspace")

    # If a file store is only used by one workspace, then we can reasonably
    # assume that all files in that store are visible to artifacts in that
    # workspace.
    for file_store in FileStore.objects.all():
        workspaces = (
            file_store.default_workspaces.all()
            | file_store.other_workspaces.all()
        )
        if workspaces.count() == 1:
            workspace = workspaces.earliest("id")
            FileInArtifact.objects.filter(
                artifact__workspace=workspace,
                file__filestore=file_store,
                complete__isnull=True,
            ).update(complete=True)

    # Everything that's left over must be considered incomplete.
    FileInArtifact.objects.filter(complete__isnull=True).update(complete=False)


def revert_file_in_artifact(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Unpopulate `FileInArtifact.complete`."""
    FileInArtifact = apps.get_model("db", "FileInArtifact")
    FileInArtifact.objects.filter(complete__isnull=False).update(complete=None)


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0074_fileinartifact_complete_initial'),
    ]

    operations = [
        migrations.RunPython(update_file_in_artifact, revert_file_in_artifact)
    ]



# Generated by Django 4.2.18 on 2025-02-09 22:45

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0125_add_group_ephemeral'),
    ]

    operations = [
        migrations.AddField(
            model_name='filestore',
            name='instance_wide',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='filestoreinscope',
            name='_file_store_instance_wide',
            field=models.BooleanField(
                db_column='file_store_instance_wide',
                default=True,
                editable=False,
                help_text='Synced from FileStore.instance_wide; do not update directly',
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='filestore',
            trigger=pgtrigger.compiler.Trigger(
                name='db_filestore_sync_instance_wide',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='UPDATE db_filestoreinscope SET file_store_instance_wide = NEW.instance_wide WHERE db_filestoreinscope.file_store_id = NEW.id AND db_filestoreinscope.file_store_instance_wide != NEW.instance_wide; RETURN NEW;',
                    hash='7dfd74c5dbb675f55873e1fa249285a172869cf3',
                    operation='INSERT OR UPDATE',
                    pgid='pgtrigger_db_filestore_sync_instance_wide_e2fab',
                    table='db_filestore',
                    when='AFTER',
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='filestoreinscope',
            trigger=pgtrigger.compiler.Trigger(
                name='db_filestoreinscope_sync_instance_wide',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='NEW.file_store_instance_wide = (SELECT db_filestore.instance_wide FROM db_filestore WHERE db_filestore.id = NEW.file_store_id); RETURN NEW;',
                    hash='04b1ed9053b9e9ea2cf78d10850742c925eeae66',
                    operation='INSERT OR UPDATE',
                    pgid='pgtrigger_db_filestoreinscope_sync_instance_wide_c82e6',
                    table='db_filestoreinscope',
                    when='BEFORE',
                ),
            ),
        ),
        migrations.AddConstraint(
            model_name='filestoreinscope',
            constraint=models.UniqueConstraint(
                condition=models.Q(('_file_store_instance_wide', False)),
                fields=('file_store',),
                name='db_filestoreinscope_unique_file_store_not_instance_wide',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the signing no-operation task."""

from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from debusine.signing.tasks import SigningNoop
from debusine.tasks import TaskConfigError


class SigningNoopTests(TestCase):
    """Test the SigningNoop task."""

    def test_execute_result_true(self) -> None:
        """If result=True, the task returns True."""
        noop = SigningNoop({"result": True})
        noop._debug_log_files_directory = TemporaryDirectory(
            prefix="debusine-tests-"
        )
        self.addCleanup(noop._debug_log_files_directory.cleanup)
        with mock.patch.object(
            noop, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertTrue(noop.execute())

    def test_execute_result_false(self) -> None:
        """If result=False, the task returns False."""
        noop = SigningNoop({"result": False})
        noop._debug_log_files_directory = TemporaryDirectory(
            prefix="debusine-tests-"
        )
        self.addCleanup(noop._debug_log_files_directory.cleanup)
        with mock.patch.object(
            noop, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertFalse(noop.execute())

    def test_no_additional_properties(self) -> None:
        """Assert no additional properties."""
        error_msg = "extra fields not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            SigningNoop({"result": "True", "input": {}})

    def test_label(self) -> None:
        """Test get_label."""
        task = SigningNoop({"result": True})
        self.assertEqual(task.get_label(), "noop")



# Generated by Django 3.2.19 on 2023-09-28 12:46

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0009_alter_token_comment'),
    ]

    operations = [
        migrations.CreateModel(
            name='Identity',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'issuer',
                    models.CharField(
                        help_text='identifier of auhoritative system for this identity',
                        max_length=512,
                    ),
                ),
                (
                    'subject',
                    models.CharField(
                        help_text='identifier of the user in the issuer system',
                        max_length=512,
                    ),
                ),
                (
                    'last_used',
                    models.DateTimeField(
                        auto_now=True,
                        help_text='last time this identity has been used',
                    ),
                ),
                (
                    'profile',
                    models.CharField(
                        blank=True,
                        help_text='URL to the user profile in the issuer system',
                        max_length=1024,
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='identities',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name='identity',
            constraint=models.UniqueConstraint(
                fields=('issuer', 'subject'),
                name='Each issuer must have unique subjects',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_workspaces."""

from datetime import timedelta
from typing import ClassVar

import yaml

from debusine.db.context import context
from debusine.db.models import Workspace
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListWorkspacesCommandTests(TabularOutputTests, TestCase):
    """Test for list_workspaces management command."""

    workspaces: ClassVar[list[Workspace]]

    @classmethod
    @context.disable_permission_checks()
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        ws1 = cls.playground.get_default_workspace()

        scope2 = cls.playground.get_or_create_scope("scope2")
        ws2 = cls.playground.create_workspace(
            scope=scope2, name="TestWorkspace2Public"
        )
        ws2.public = True
        ws2.save()

        ws3 = cls.playground.create_workspace(
            scope=scope2, name="TestWorkspace3WithOtherWorkspaces"
        )

        ws4 = cls.playground.create_workspace(
            scope=scope2, name="TestWorkspace4WithExpiration"
        )
        ws4.default_expiration_delay = timedelta(days=400)
        ws4.save()

        cls.workspaces = [ws1, ws2, ws3, ws4]

    def test_list_workspaces(self) -> None:
        """List workspace command prints workspace information."""
        with self.assertPrintsTable() as output:
            stdout, stderr, _ = call_command('list_workspaces')

        def _format_expiration(ws: Workspace) -> str:
            if days := ws.default_expiration_delay.days:
                return str(days)
            return "Never"

        self.assertEqual(output.col(0), [str(ws) for ws in self.workspaces])
        self.assertEqual(
            output.col(1), [str(ws.public) for ws in self.workspaces]
        )
        self.assertEqual(
            output.col(2),
            [_format_expiration(ws) for ws in self.workspaces],
        )

    def test_list_workspaces_yaml(self) -> None:
        """List workspace command prints yaml workspace information."""
        stdout, stderr, _ = call_command('list_workspaces', "--yaml")
        data = yaml.safe_load(stdout)
        self.assertEqual(
            data,
            [
                {
                    'expiration': 'Never',
                    'name': 'debusine/System',
                    'public': True,
                },
                {
                    'expiration': 'Never',
                    'name': 'scope2/TestWorkspace2Public',
                    'public': True,
                },
                {
                    'expiration': 'Never',
                    'name': 'scope2/TestWorkspace3WithOtherWorkspaces',
                    'public': False,
                },
                {
                    'expiration': 400,
                    'name': 'scope2/TestWorkspace4WithExpiration',
                    'public': False,
                },
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Fix WorkRequest.task_type."""

from functools import partial

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from debusine.tasks.models import TaskTypes


def fix_task_type(
    apps: StateApps,
    schema_editor: BaseDatabaseSchemaEditor,
    *,
    reverse: bool = False,
) -> None:
    """
    Fix WorkRequest.task_type.

    From 0027_workrequest_task_type until 0030_move_task_type, this field
    used lower-case database values and a lower-case default.  These are no
    longer valid.
    """
    WorkRequest = apps.get_model("db", "WorkRequest")
    for work_request in WorkRequest.objects.all():
        if reverse:
            work_request.task_type = work_request.task_type.lower()
        else:
            work_request.task_type = TaskTypes[work_request.task_type.upper()]
        work_request.save()


class Migration(migrations.Migration):

    dependencies = [
        ("db", "0035_autopkgtest_extra_environment"),
    ]

    operations = [
        migrations.RunPython(
            fix_task_type, partial(fix_task_type, reverse=True)
        )
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for table columns."""

from unittest import mock

from django.test import RequestFactory

from debusine.db.models import User
from debusine.test.django import TestCase
from debusine.web.views.table import (
    Column,
    DateTimeColumn,
    NumberColumn,
    Ordering,
    Sort,
    StringColumn,
    Table,
)
from debusine.web.views.table.columns import BoundColumn, Columns


class BoundColumnTests(TestCase):
    """Tests for :py:class:`BoundColumn`."""

    def _column(self, column: Column) -> BoundColumn[User]:
        """Create a column bound to a mock table."""

        class _Table(Table[User]):
            def init(self) -> None:
                super().init()
                self.add_column(column.name, column)

        table = _Table(
            RequestFactory().get("/"),
            User.objects.all(),
            default_order=column.name,
        )

        return table.columns[column.name]

    def assertOrderings(self, col: BoundColumn[User], names: list[str]) -> None:
        """Ensure the orderings match the given "what.how" names."""
        self.assertEqual([f"{o.what}.{o.how}" for o in col.orderings], names)

    def test_column(self) -> None:
        """Test Column defaulting to StringColumn."""
        col = self._column(Column("User", ordering="username", name="user"))
        self.assertOrderings(col, ["user.asc", "user.desc"])
        self.assertEqual(str(col), "user")
        self.assertEqual(repr(col), "BoundColumn('user')")
        self.assertEqual(col.name, "user")
        self.assertEqual(col.title, "User")
        ctx = col.get_context_data()
        self.assertEqual(ctx["column"], col)
        self.assertHTMLValid(col.sort_icon)

    def test_string_column(self) -> None:
        """Test StringColumn."""
        col = self._column(
            StringColumn("Email", ordering="email", name="email")
        )
        self.assertOrderings(col, ["email.asc", "email.desc"])
        self.assertEqual(str(col), "email")
        self.assertEqual(repr(col), "BoundColumn('email')")
        self.assertEqual(col.name, "email")
        self.assertEqual(col.title, "Email")
        ctx = col.get_context_data()
        self.assertEqual(ctx["column"], col)
        self.assertHTMLValid(col.sort_icon)

    def test_number_column(self) -> None:
        """Test NumberColumn."""
        col = self._column(NumberColumn("Age", ordering="age", name="age"))
        self.assertOrderings(col, ["age.asc", "age.desc"])
        self.assertEqual(str(col), "age")
        self.assertEqual(repr(col), "BoundColumn('age')")
        self.assertEqual(col.name, "age")
        self.assertEqual(col.title, "Age")
        self.assertEqual(col.get_context_data(), {"column": col})
        self.assertHTMLValid(col.sort_icon)

    def test_datetime_column(self) -> None:
        """Test DateTimeColumn."""
        col = self._column(
            DateTimeColumn("Last Seen", ordering="last_seen", name="last_seen")
        )
        self.assertOrderings(col, ["last_seen.asc", "last_seen.desc"])
        self.assertEqual(str(col), "last_seen")
        self.assertEqual(repr(col), "BoundColumn('last_seen')")
        self.assertEqual(col.name, "last_seen")
        self.assertEqual(col.title, "Last Seen")
        self.assertEqual(col.get_context_data(), {"column": col})
        self.assertHTMLValid(col.sort_icon)

    def test_custom_sort(self) -> None:
        """Test DateTimeColumn."""
        col = self._column(
            Column(
                "Test",
                ordering=Ordering(foo=Sort("username", "Foo", "icon")),
                name="test",
            )
        )
        self.assertOrderings(col, ["test.foo"])
        self.assertEqual(str(col), "test")
        self.assertEqual(repr(col), "BoundColumn('test')")
        self.assertEqual(col.name, "test")
        self.assertEqual(col.title, "Test")
        self.assertEqual(col.get_context_data(), {"column": col})
        self.assertHTMLValid(col.sort_icon)

    def test_unsortable_column(self) -> None:
        """Test data of unsortable columns."""
        col = self._column(Column("Extra", name="extra"))
        self.assertEqual(col.orderings, [])
        self.assertEqual(str(col), "extra")
        self.assertEqual(repr(col), "BoundColumn('extra')")
        self.assertEqual(col.name, "extra")
        self.assertEqual(col.title, "Extra")
        self.assertEqual(col.get_context_data(), {"column": col})
        self.assertEqual(col.sort_icon, "")


class ColumnsTests(TestCase):
    """Tests for :py:class:`Columns`."""

    def _columns(self) -> Columns[User]:
        """Create a Columns bound to a mock table."""
        table = mock.Mock()
        table.request = RequestFactory().get("/")
        return Columns(table)

    def test_init(self) -> None:
        columns = self._columns()
        self.assertEqual(columns.columns, {})

    def test_add(self) -> None:
        columns = self._columns()
        col = columns.add(
            "user", Column("User", ordering="username", name="user")
        )
        self.assertIs(col.table, columns.table)
        self.assertEqual(columns.columns, {"user": col})

    def test_getitem(self) -> None:
        columns = self._columns()
        col = columns.add(
            "user", Column("User", ordering="username", name="user")
        )
        self.assertIs(columns["user"], col)
        with self.assertRaises(KeyError):
            columns["foo"]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for asset data."""

import abc
from enum import StrEnum
from typing import Annotated, Any, Literal, Union

from debusine.utils import DjangoChoicesEnum

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic as pydantic  # type: ignore


class AssetCategory(DjangoChoicesEnum):
    """Possible asset categories."""

    CLOUD_PROVIDER_ACCOUNT = "debusine:cloud-provider-account"
    SIGNING_KEY = "debusine:signing-key"


class BaseAssetDataModel(pydantic.BaseModel, abc.ABC):
    """Base pydantic model for asset data and their components."""

    class Config:
        """Set up stricter pydantic Config."""

        validate_assignment = True
        extra = pydantic.Extra.forbid


class ModelCategory(BaseAssetDataModel):
    """All models for a category, organized by discriminator value."""

    discriminator: str | None = None
    models: dict[str | None, type[BaseAssetDataModel]] = {}


_data_models: dict[AssetCategory, ModelCategory] = {}


class BaseRegisteredAssetDataModel(BaseAssetDataModel, abc.ABC):
    """An asset data model registered for use by `asset_data_model`."""

    def __init_subclass__(
        cls,
        category: AssetCategory,
        discriminator: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Register subclass in _data_models."""
        super().__init_subclass__(**kwargs)
        if category not in _data_models:
            _data_models[category] = ModelCategory(discriminator=discriminator)
        modelcategory = _data_models[category]
        assert modelcategory.discriminator == discriminator
        discriminator_value: str | None = None
        if discriminator:
            discriminator_value = cls.__fields__[discriminator].default
        modelcategory.models[discriminator_value] = cls


def asset_data_model(
    category: AssetCategory | str,
    data: dict[str, Any],
) -> BaseAssetDataModel:
    """Load the DataModel for an asset, with specified category and data."""
    category = AssetCategory(category)
    if category not in _data_models:
        raise ValueError(f"No data model for {category} exists.")
    modelcategory = _data_models[category]
    if modelcategory.discriminator is None:
        model = modelcategory.models[None]
    else:
        if modelcategory.discriminator not in data:
            raise ValueError(
                f"Discriminator key {modelcategory.discriminator} "
                f"required for {category} missing in data."
            )
        discriminator_value = data[modelcategory.discriminator]
        if discriminator_value not in modelcategory.models:
            raise ValueError(
                f"No data model for {category} with "
                f"{modelcategory.discriminator}={discriminator_value}."
            )
        model = modelcategory.models[discriminator_value]
    return model.parse_obj(data)


class KeyPurpose(DjangoChoicesEnum):
    """Choices for SigningKey.purpose."""

    UEFI = "uefi"
    OPENPGP = "openpgp"


class SigningKeyData(
    BaseRegisteredAssetDataModel, category=AssetCategory.SIGNING_KEY
):
    """Data for a debusine:signing-key asset."""

    purpose: KeyPurpose
    fingerprint: str
    public_key: str
    description: str | None


class CloudProvidersType(StrEnum):
    """Choices for CloudProviderAccountData.provider_type."""

    AWS = "aws"
    HETZNER = "hetzner"
    DUMMY = "dummy"


class AWSProviderAccountConfiguration(BaseAssetDataModel):
    """Non-secret configuration for an AWS account."""

    region_name: str | None = None
    ec2_endpoint_url: str | None = None
    s3_endpoint_url: str | None = None


class AWSProviderAccountCredentials(BaseAssetDataModel):
    """Secret credentials for an AWS account."""

    access_key_id: str
    secret_access_key: str


class AWSProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for AWS."""

    provider_type: Literal[CloudProvidersType.AWS] = CloudProvidersType.AWS
    name: str
    configuration: AWSProviderAccountConfiguration
    credentials: AWSProviderAccountCredentials


class HetznerProviderAccountConfiguration(BaseAssetDataModel):
    """Non-secret configuration for a Hetzner Cloud account."""

    region_name: str


class HetznerProviderAccountCredentials(BaseAssetDataModel):
    """Secret credentials for a Hetzner Cloud account."""

    api_token: str


class HetznerProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for Hetzner Cloud."""

    provider_type: Literal[CloudProvidersType.HETZNER] = (
        CloudProvidersType.HETZNER
    )
    name: str
    configuration: HetznerProviderAccountConfiguration
    credentials: HetznerProviderAccountCredentials


class DummyProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for test cases."""

    provider_type: Literal[CloudProvidersType.DUMMY] = CloudProvidersType.DUMMY
    name: str
    secret: str = "secret"


CloudProviderAccountData = Annotated[
    Union[
        AWSProviderAccountData,
        HetznerProviderAccountData,
        DummyProviderAccountData,
    ],
    pydantic.Field(discriminator="provider_type"),
]



# Generated by Django 4.2.19 on 2025-04-11 09:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0141_add_clientenroll'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='workflows_need_update',
            field=models.BooleanField(default=False),
        ),
        migrations.AddIndex(
            model_name='workrequest',
            index=models.Index(
                models.F('workflows_need_update'),
                name='db_workrequest_wf_update_idx',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command workspace."""
import io
from datetime import timedelta
from typing import ClassVar

import yaml
from django.core.management import CommandError

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.context import context
from debusine.db.models import (
    ArtifactRelation,
    Scope,
    WorkflowTemplate,
    Workspace,
    default_workspace,
)
from debusine.db.models.workspaces import WorkspaceRole
from debusine.db.playground import Playground, scenarios
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class WorkspaceCommandsTests(TabularOutputTests, TestCase):
    """Tests for workspace management commands."""

    scope: ClassVar[Scope]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common test data."""
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope(name="localtest")

    def test_define_scenario(self) -> None:
        """Test define command."""
        # define/create, with errors
        with self.assertRaisesRegex(
            CommandError,
            "Error: owners_group is required when creating workspace",
        ):
            call_command("workspace", "create", "localtest/myws")

        with self.assertRaisesRegex(
            CommandError,
            "scope_workspace 'myws' should be in the form"
            " 'scopename/workspacename'",
        ):
            call_command("workspace", "create", "myws", "Owners-myws")

        with self.assertRaisesRegex(
            CommandError,
            "Scope 'nonexistent' not found",
        ):
            call_command(
                "workspace", "create", "nonexistent/my ws", "Owners-myws"
            )

        with self.assertRaisesRegex(
            CommandError,
            "Error creating workspace: 'my ws' is not a valid workspace name",
        ):
            call_command(
                "workspace", "create", "localtest/my ws", "Owners-myws"
            )

        with self.assertRaisesRegex(
            CommandError, "Group 'Owners-myws' not found in scope 'localtest'"
        ):
            call_command("workspace", "create", "localtest/myws", "Owners-myws")

        self.playground.create_group("Owners-myws", scope=self.scope)
        self.playground.create_group("Owners-myws2", scope=self.scope)

        # create/define, idempotent
        for i in (1, 2):
            stdout, stderr, exit_code = call_command(
                "workspace", "define", "localtest/myws", "Owners-myws"
            )
            # assert workspace and role
            self.assertEqual(stdout, "")
            self.assertEqual(stderr, "")
            self.assertEqual(exit_code, 0)
            workspace = Workspace.objects.get(scope=self.scope, name="myws")
            role = WorkspaceRole.objects.get(
                resource=workspace, group__name="Owners-myws"
            )
            self.assertEqual(role.role, WorkspaceRole.Roles.OWNER)
            self.assertFalse(workspace.public)
            self.assertEqual(workspace.default_expiration_delay, timedelta(30))
            self.assertQuerySetEqual(
                workspace.collections.values_list("category", flat=True),
                SINGLETON_COLLECTION_CATEGORIES,
                ordered=False,
            )

        # Set parameters
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
            "Owners-myws",
            "--public",
            "--default-expiration-delay",
            "0",
            "--no-singleton-collections",
        )
        workspace = Workspace.objects.get(scope=self.scope, name="myws-new")
        self.assertTrue(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, timedelta(0))
        self.assertQuerySetEqual(
            workspace.collections.values_list("category", flat=True), []
        )

        # Change parameters
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
            "Owners-myws2",
            "--private",
            "--default-expiration-delay",
            "10",
        )
        workspace.refresh_from_db()
        self.assertFalse(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, timedelta(10))
        role = WorkspaceRole.objects.get(
            resource=workspace, group__name="Owners-myws2"
        )
        self.assertEqual(role.role, WorkspaceRole.Roles.OWNER)

        # Don't reset parameters to defaults
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
        )
        self.assertFalse(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, timedelta(10))

    def test_rename_scenario(self) -> None:
        """Test rename command."""
        self.playground.create_group("Owners-myws", scope=self.scope)
        call_command("workspace", "create", "localtest/myws", "Owners-myws")

        # rename, with errors
        with self.assertRaisesRegex(
            CommandError, "Workspace 'toto' not found in scope 'localtest'"
        ):
            call_command("workspace", "rename", "localtest/toto", "toto2")

        stdout, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "my ws"
        )
        self.assertEqual(
            stderr,
            """Renamed workspace would be invalid:
* name: 'my ws' is not a valid workspace name
""",
        )
        self.assertEqual(exit_code, 3)
        self.assertTrue(
            Workspace.objects.filter(scope=self.scope, name="myws").exists()
        )

        call_command(
            "workspace", "create", "localtest/myws-exists", "Owners-myws"
        )
        stdout, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "myws-exists"
        )
        self.assertEqual(
            stderr,
            """Renamed workspace would be invalid:
* __all__: Workspace with this Scope and Name already exists.
""",
        )
        self.assertEqual(exit_code, 3)

        # rename, idempotent
        stdout, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "myws-renamed"
        )
        self.assertEqual(exit_code, 0)
        self.assertFalse(
            Workspace.objects.filter(scope=self.scope, name="myws").exists()
        )
        self.assertTrue(
            Workspace.objects.filter(
                scope=self.scope, name="myws-renamed"
            ).exists()
        )

        stdout, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws-renamed", "myws-renamed"
        )
        self.assertEqual(exit_code, 0)

    def test_list(self) -> None:
        """Test list command."""
        # Setup
        self.playground.create_group("Owners-myws", scope=self.scope)

        call_command(
            "workspace", "create", "localtest/public", "Owners-myws", "--public"
        )
        call_command("workspace", "create", "localtest/otherfs", "Owners-myws")
        call_command(
            "workspace",
            "create",
            "localtest/noexpire",
            "Owners-myws",
            "--default-expiration-delay",
            "0",
        )

        # list
        with self.assertPrintsTable() as output:
            stdout, stderr, _ = call_command("workspace", "list")

        self.assertEqual(
            output.col(0),
            [
                "debusine/System",
                "localtest/public",
                "localtest/otherfs",
                "localtest/noexpire",
            ],
        )
        self.assertEqual(output.col(1), ["True", "True", "False", "False"])
        self.assertEqual(output.col(2), ["Never", "30", "30", "Never"])

        # --yaml
        stdout, stderr, _ = call_command("workspace", "list", "--yaml")
        data = yaml.safe_load(stdout)
        expected_data = [
            {
                'expiration': 'Never',
                'name': 'debusine/System',
                'public': True,
            },
            {
                'expiration': 30,
                'name': 'localtest/public',
                'public': True,
            },
            {
                'expiration': 30,
                'name': 'localtest/otherfs',
                'public': False,
            },
            {
                'expiration': 'Never',
                'name': 'localtest/noexpire',
                'public': False,
            },
        ]
        self.assertEqual(data, expected_data)

        # scope
        stdout, stderr, _ = call_command(
            "workspace", "list", "localtest", "--yaml"
        )
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data[1:])

    def test_roles(self) -> None:
        """Test roles commands."""
        self.playground.create_group("Admins", scope=self.scope)
        self.playground.create_group("Employees", scope=self.scope)
        self.playground.create_group("Contractors", scope=self.scope)

        # fmt: off
        call_command("workspace", "create", "localtest/myws", "Admins")

        with self.assertRaisesRegex(
            CommandError,
            "Error assigning role: 'xxx' is not a valid WorkspaceRoles",
        ):
            call_command(
                "workspace", "grant_role", "localtest/myws",
                "xxx", "Employees",
            )

        call_command(
            "workspace", "grant_role", "localtest/myws",
            "contributor", "Employees", "Contractors",
        )

        with self.assertPrintsTable() as output:
            stdout, stderr, exit_code = call_command(
                "workspace", "list_roles", "localtest/myws"
            )
        self.assertEqual(output.col(0), ["Admins", "Contractors", "Employees"])
        self.assertEqual(output.col(1), ["owner", "contributor", "contributor"])
        self.assertEqual(exit_code, 0)

        stdout, stderr, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        expected_data = [
            {'group': 'Admins', 'role': 'owner'},
            {'group': 'Contractors', 'role': 'contributor'},
            {'group': 'Employees', 'role': 'contributor'},
        ]
        self.assertEqual(data, expected_data)
        self.assertEqual(exit_code, 0)

        # idempotent
        call_command(
            "workspace", "grant_role", "localtest/myws",
            "contributor", "Employees", "Contractors",
        )
        stdout, stderr, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data)

        # revoke_role
        stdout, stderr, exit_code = call_command(
            "workspace", "revoke_role", "localtest/myws",
            "contributor", "Employees"
        )
        stdout, stderr, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data[0:-1])
        # fmt: on

    @context.disable_permission_checks()
    def create_workspace_for_delete(self) -> Workspace:
        """Create a test workspace."""
        scope = self.playground.get_or_create_scope("scope")
        workspace = self.playground.create_workspace(name="Test", scope=scope)
        self.playground.create_group_role(
            workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.playground.get_default_user()],
        )
        for architecture in ("amd64", "s390x"):
            self.playground.create_debian_environment(
                codename="bookworm",
                architecture=architecture,
                workspace=workspace,
            )
        work_request = self.playground.create_work_request(
            workspace=workspace, task_name="noop"
        )
        collection = self.playground.create_collection(
            "test", CollectionCategory.WORKFLOW_INTERNAL, workspace=workspace
        )
        artifact_hello, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=[
                "hello_1.0-1.dsc",
                "hello_1.0-1.debian.tar.xz",
                "hello_1.0.orig.tar.xz",
            ],
            workspace=workspace,
            create_files=True,
            skip_add_files_in_store=True,
        )
        artifact_hello.created_by_work_request = work_request
        artifact_hello.save()

        artifact_hello_traditional, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello-traditional",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=[
                "hello-traditional.0-1.dsc",
                "hello-traditional.0-1.debian.tar.xz",
                "hello-traditional.0.orig.tar.xz",
            ],
            workspace=workspace,
            create_files=True,
            skip_add_files_in_store=True,
        )
        self.playground.create_artifact_relation(
            artifact_hello_traditional,
            artifact_hello,
            ArtifactRelation.Relations.RELATES_TO,
        )
        collection.manager.add_artifact(
            artifact_hello,
            user=self.playground.get_default_user(),
            name="hello",
        )
        collection.manager.add_artifact(
            artifact_hello_traditional,
            user=self.playground.get_default_user(),
            name="hello-traditional",
        )
        WorkflowTemplate.objects.create(
            name="test", workspace=workspace, task_name="noop"
        )

        sbuild_template = self.playground.create_workflow_template(
            name="Build package",
            task_name="sbuild",
            task_data={},
            workspace=workspace,
        )

        udev = self.playground.create_source_artifact(
            name="udev",
            version="252.26-1~deb12u2",
            create_files=True,
            workspace=workspace,
        )

        workflow = self.playground.create_workflow(
            sbuild_template,
            task_data={
                "input": {
                    "source_artifact": udev.pk,
                },
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64", "s390x"],
            },
        )
        workflow.mark_running()
        workflow.save()

        assert workflow.internal_collection is not None
        self.assertEqual(workflow.internal_collection.workflow, workflow)

        # TODO: add more kinds of elements to the workspace, to make
        # sure deletion catches them
        return workspace

    def test_delete_workspace(self) -> None:
        """Delete an existing workspace."""
        workspace = self.create_workspace_for_delete()

        stdout, stderr, _ = call_command(
            "workspace",
            "delete",
            "scope/Test",
            "--yes",
        )
        self.assertEqual("", stdout)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(id=workspace.id)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(name="Test")

    def test_delete_workspace_playground_ui_scenario(self) -> None:
        """Delete a workspace with the playground UI scenario."""
        playground = Playground(default_workspace_name="Playground")
        scenario = scenarios.UIPlayground()
        playground.build_scenario(scenario)

        stdout, stderr, _ = call_command(
            "workspace",
            "delete",
            str(scenario.workspace),
            "--yes",
        )
        self.assertEqual("", stdout)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(id=scenario.workspace.id)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(
                scope=scenario.scope, name=scenario.workspace.name
            )

    def test_delete_workspace_confirmation(self) -> None:
        """delete_workspace doesn't delete (user does not confirm)."""
        workspace = self.create_workspace_for_delete()

        call_command(
            "workspace",
            "delete",
            "scope/Test",
            stdin=io.StringIO("N\n"),
        )
        self.assertQuerySetEqual(
            Workspace.objects.filter(name="Test"), [workspace]
        )

        call_command(
            "workspace",
            "delete",
            "scope/Test",
            stdin=io.StringIO("\n"),
        )
        self.assertQuerySetEqual(
            Workspace.objects.filter(name="Test"), [workspace]
        )

    def test_delete_default_workspace(self) -> None:
        """Default workspace does not get deleted."""
        workspace = default_workspace()
        with self.assertRaisesRegex(
            CommandError,
            fr"^Workspace {workspace.scope.name}/{workspace.name}"
            " cannot be deleted$",
        ) as exc:
            call_command(
                "workspace",
                "delete",
                str(workspace),
                "--yes",
            )
        self.assertEqual(exc.exception.returncode, 3)



# Generated by Django 3.2.19 on 2024-03-27 15:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0027_workrequest_task_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='dependencies',
            field=models.ManyToManyField(
                related_name='reverse_dependencies', to='db.WorkRequest'
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='event_reactions',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='parent',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to='db.workrequest',
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='unblock_strategy',
            field=models.CharField(
                choices=[
                    ('deps', 'Dependencies have completed'),
                    ('manual', 'Manually unblocked'),
                ],
                default='deps',
                editable=False,
                max_length=6,
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='workflow_data',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='status',
            field=models.CharField(
                choices=[
                    ('pending', 'Pending'),
                    ('running', 'Running'),
                    ('completed', 'Completed'),
                    ('aborted', 'Aborted'),
                    ('blocked', 'Blocked'),
                ],
                default='pending',
                editable=False,
                max_length=9,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for workflows."""

from debusine.db.models import WorkRequest
from debusine.server.tasks.tests.helpers import TestBaseWaitTask
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData, TaskTypes
from debusine.web.templatetags.workflows import workflow_runtime_status_small

# templatetags loading code will attempt to import tests, debusine.test
# dependencies (such as lxml) may not be available at runtime
try:
    from debusine.test.django import TestCase
except ImportError:
    from unittest import TestCase  # type: ignore[assignment]


class WorkflowRuntimeStatusSmallTests(TestCase):
    """Tests for workflow_runtime_status_small tag."""

    def setUp(self) -> None:
        """Set up test."""
        super().setUp()
        self.workflow = self.playground.create_workflow()

    def test_status_is_completed(self) -> None:
        """Test workflow status is completed."""
        self.workflow.status = WorkRequest.Statuses.COMPLETED
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.COMPLETED
        )

        actual = workflow_runtime_status_small(self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span class="badge text-bg-primary" title="Completed">C</span>',
        )

    def test_status_is_running(self) -> None:
        """Test workflow status is running and runtime status is blocked."""
        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.BLOCKED
        )

        actual = workflow_runtime_status_small(self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span class="badge text-bg-secondary" title="Running">R</span>'
            '<span class="badge text-bg-dark" title="Blocked">B</span>',
        )

    def test_runtime_status_is_needs_input(self) -> None:
        """Test workflow runtime status is needs input."""
        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )

        class WaitNoop(
            TestBaseWaitTask[BaseTaskData, BaseDynamicTaskData],
        ):
            TASK_VERSION = 1

            def _execute(self) -> bool:
                raise NotImplementedError()

        self.workflow.create_child(
            "noop",
        )

        children_needs_input = self.workflow.create_child(
            "waitnoop",
            task_type=TaskTypes.WAIT,
            status=WorkRequest.Statuses.RUNNING,
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )

        self.workflow.create_child(
            "waitnoop",
            task_type=TaskTypes.WAIT,
            status=WorkRequest.Statuses.RUNNING,
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )

        self.workflow.create_child("noop", status=WorkRequest.Statuses.ABORTED)

        actual = workflow_runtime_status_small(self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span title="Running" class="badge text-bg-secondary">R</span> '
            f'<a href="{children_needs_input.get_absolute_url()}">'
            '<span title="Needs Input"class="badge text-bg-secondary">I</span>'
            '</a>',
        )

    def test_runtime_status_is_needs_input_no_work_request(self) -> None:
        """
        Test workflow runtime status is needs input but no work request waiting.

        The workflow was NEEDS_INPUT but no work request is waiting (finished
        before getting the URL).
        """
        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )

        actual = workflow_runtime_status_small(self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span title="Running" class="badge text-bg-secondary">R</span> '
            '<span title="Needs Input" class="badge text-bg-secondary">I'
            '</span>',
        )



# Generated by Django 4.2.15 on 2024-12-16 10:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0108_fill_scope_label'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scope',
            name='label',
            field=models.CharField(
                help_text='User-visible name for the scope',
                max_length=255,
                unique=True,
            ),
        ),
    ]
